"""Base class for test cases in Django-based applications."""

from __future__ import annotations

import inspect
import os
import re
import socket
import sys
import threading
import warnings
from collections import OrderedDict
from contextlib import contextmanager
from importlib import import_module
from importlib.machinery import ModuleSpec
from importlib.util import module_from_spec
from typing import (Any, Dict, Iterator, List, Optional, Sequence,
                    TYPE_CHECKING, Type, Union)
from unittest.util import safe_repr

from django.apps import apps
from django.conf import settings
from django.core import serializers
from django.core.exceptions import ValidationError
from django.core.handlers.wsgi import WSGIHandler
from django.core.management import call_command
from django.core.servers import basehttp
from django.db import (DatabaseError, DEFAULT_DB_ALIAS, IntegrityError,
                       connections, router)
from django.db.models import Model
from django.template import Node
from django.test import testcases
from typing_extensions import NotRequired, TypedDict

try:
    from django_evolution.models import Evolution, Version
except ImportError:
    Evolution = None
    Version = None

from djblets.db.query_comparator import compare_queries
from djblets.siteconfig.models import SiteConfiguration

if TYPE_CHECKING:
    from djblets.db.query_comparator import (CompareQueriesContext,
                                             ExpectedQuery,
                                             QueryMismatchedAttr)


class StubNodeList(Node):
    def __init__(self, default_text):
        self.default_text = default_text

    def render(self, context):
        return self.default_text


class StubParser:
    def __init__(self, default_text):
        self.default_text = default_text

    def parse(self, until):
        return StubNodeList(self.default_text)

    def delete_first_token(self):
        pass


class ExpectedWarning(TypedDict):
    """An expected warning from an assertion.

    This is used for :py:meth:`TestCase.assertWarnings`.

    Version Added:
        3.2
    """

    #: The expected class for the warning.
    #:
    #: Type:
    #:     type
    cls: Type[Warning]

    #: The expected message for the warning.
    #:
    #: If not provided, messages won't be compared.
    #:
    #: Type:
    #:     str
    message: NotRequired[Optional[str]]


class TestCase(testcases.TestCase):
    """Base class for test cases.

    Individual tests on this TestCase can use the :py:func:`add_fixtures`
    decorator to add or replace the fixtures used for the test.
    """
    ws_re = re.compile(r'\s+')

    def __call__(self, *args, **kwargs):
        method = getattr(self, self._testMethodName)
        old_fixtures = getattr(self, 'fixtures', None)

        if hasattr(method, '_fixtures'):
            if (getattr(method, '_replace_fixtures') or
                old_fixtures is None):
                self.fixtures = method._fixtures
            else:
                self.fixtures = old_fixtures + method._fixtures

        super(TestCase, self).__call__(*args, **kwargs)

        self.fixtures = old_fixtures

    def shortDescription(self):
        """Returns the description of the current test.

        This changes the default behavior to replace all newlines with spaces,
        allowing a test description to span lines. It should still be kept
        short, though.
        """
        doc = self._testMethodDoc

        if doc is not None:
            doc = doc.split('\n\n', 1)[0]
            doc = self.ws_re.sub(' ', doc).strip()

        return doc

    @contextmanager
    def siteconfig_settings(self, settings):
        """Temporarily sets siteconfig settings for a test.

        Subclasses should override this if they want to run a method like
        :py:func:`~djblets.siteconfig.django_settings.apply_django_settings`
        before and after each test.

        Args:
            settings (dict):
                The new siteconfig settings to set.

        Context:
            The current site configuration will contain the new settings for
            this test.
        """
        siteconfig = SiteConfiguration.objects.get_current()

        old_settings = {}

        for key, value in settings.items():
            old_settings[key] = siteconfig.get(key)
            siteconfig.set(key, value)

        siteconfig.save()

        try:
            yield
        finally:
            for key, value in old_settings.items():
                siteconfig.set(key, value)

            siteconfig.save()

    def assertAttrsEqual(self, obj, attrs, msg=None):
        """Assert that attributes on an object match expected values.

        This will compare each attribute defined in ``attrs`` against the
        corresponding attribute on ``obj``. If the attribute value does not
        match, or the attribute is missing, this will assert.

        Args:
            obj (object):
                The object to compare attribute values to.

            attrs (dict):
                A dictionary of expected attribute names and values.

            msg (unicode, optional):
                A custom message to include if there's a failure.

        Raises:
            AssertionError:
                An attribute was not found or the value did not match.
        """
        for key, value in attrs.items():
            try:
                attr_value = getattr(obj, key)

                if attr_value != value:
                    raise self.failureException(self._formatMessage(
                        msg,
                        '%s: %s != %s'
                        % (key, safe_repr(attr_value),
                           safe_repr(value))))
            except AttributeError:
                raise self.failureException(self._formatMessage(
                    msg,
                    'Attribute "%s" was not found on %s'
                    % (key, safe_repr(obj))))

    def assertRaisesValidationError(self, expected_messages, *args, **kwargs):
        """Assert that a ValidationError is raised with the given message(s).

        This is a wrapper around :py:meth:`assertRaisesMessage` with a
        :py:class:`ValidationError` that handles converting the expected
        messages into a list (if it isn't already) and then converting that
        into a string representation, which is what
        :py:meth:`assertRaisesMessage` will be checking against.

        Args:
            expected_messages (list or unicode):
                The expected messages as either a list of strings or a
                single string.

            args:
                Additional arguments to pass to :py:meth:`assertRaisesMessage`.

            kwargs:
                Additional keyword arguments to pass to
                :py:meth:`assertRaisesMessage`.
        """
        if isinstance(expected_messages, str):
            expected_messages = [expected_messages]

        return self.assertRaisesMessage(ValidationError,
                                        repr(expected_messages),
                                        *args,
                                        **kwargs)

    def assertRaisesMessage(self, expected_exception, expected_message,
                            *args, **kwargs):
        """Assert that an exception is raised with a given message.

        This is a replacement for Django's assertRaisesMessage that behaves
        well with a design change in Python 2.7.9/10, without crashing.
        """
        # The difference between this version and Django's is that we're
        # not taking the callable_obj as an argument with a default value and
        # passing it down to assertRaisesRegex. Python 2.7.9/10's
        # implementation defaults callable_obj to a special value, which
        # Django overrides.
        return self.assertRaisesRegex(expected_exception,
                                      re.escape(expected_message),
                                      *args, **kwargs)

    # NOTE: This overrides TestCase.assertWarns (introducecd in Python 3.2),
    #       adding 'message' and removing a callable argument and custom
    #       error message. It's not compatible with that function. We should
    #       consider this for future deprecation.
    @contextmanager
    def assertWarns(
        self,
        cls: Type[Warning] = DeprecationWarning,
        message: Optional[str] = None,
    ) -> Iterator[None]:
        """Assert that a warning is generated with a given message.

        This method only supports code which generates a single warning.
        Tests which make use of code generating multiple warnings will
        need to manually catch their warnings.

        Args:
            cls (type, optional):
                The expected warning type.

                If not provided, this defaults to a
                :py:exc:`DeprecationWarning`.

            message (unicode, optional):
                The expected error message, if any.

        Context:
            The test to run.
        """
        warning_list: List[ExpectedWarning] = [
            {
                'cls': cls,
                'message': message,
            },
        ]

        with self.assertWarnings(warning_list):
            yield

    @contextmanager
    def assertWarnings(
        self,
        warning_list: List[ExpectedWarning],
    ) -> Iterator[None]:
        """Assert that multiple warnings were generated.

        This method accepts a sequence of warnings that must be matched in
        order. Each item must be an instance of a warning with a message.
        The type and messages of the warnings must match.

        Version Added:
            3.2

        Args:
            warning_list (list of dict, optional):
                The list of expected warnings, in order.

                Each item is a dictionary in the format described in
                :py:class:`ExpectedWarning`.

        Context:
            The test to run.
        """
        with warnings.catch_warnings(record=True) as w:
            # Some warnings such as DeprecationWarning are filtered by
            # default, stop filtering them.
            warnings.simplefilter('always')

            # We do need to ignore this one, though, or a lot of things will
            # fail. This is specific to Python 3.10 and the versions of six
            # in Django.
            warnings.filterwarnings(
                'ignore',
                r'_SixMetaPathImporter.find_spec\(\) not found')

            self.assertEqual(len(w), 0)

            try:
                yield
            finally:
                self.assertEqual(len(w), len(warning_list))

                for i, (emitted, expected) in enumerate(zip(w, warning_list),
                                                        start=1):
                    expected_cls = expected['cls']
                    expected_message = expected.get('message')

                    self.assertTrue(
                        issubclass(emitted.category, expected_cls),
                        'Warning #%d: Class %r != %r'
                        % (i, emitted.category, expected_cls))

                    if expected_message is not None:
                        self.assertEqual(
                            str(emitted.message),
                            expected_message,
                            'Warning #%d: Message %r != %r'
                            % (i, str(emitted.message), str(expected_message)))

    @contextmanager
    def assertNoWarnings(self):
        """Assert that a warning is not generated.

        Context:
            The test to run.
        """
        with self.assertWarnings([]):
            yield

    @contextmanager
    def assertQueries(
        self,
        queries: Sequence[Union[ExpectedQuery,
                                Dict[str, Any]]],
        num_statements: Optional[int] = None,
        *,
        with_tracebacks: bool = False,
        traceback_size: int = 15,
        check_join_types: bool = True,
        check_subqueries: bool = True,
    ) -> Iterator[None]:
        """Assert the number and complexity of queries.

        This provides advanced checking of queries, allowing the caller to
        match filtering, JOINs, ordering, selected fields, and more.

        This takes a list of dictionaries with query information. Each
        contains the keys in :py:class:`ExpectedQuery`.

        Version Changed:
            5.0:
            * Turned on ``check_subqueries`` and ``check_join_types`` by
              default.

        Version Changed:
            3.4:
            * Added ``with_tracebacks``, ``tracebacks_size``,
              ``check_join_types``, and ``check_subqueries`` arguments.
            * Added support for type hints for expected queries.
            * Query output can now show notes (when populating
              :py:attr:`ExpectedQuery.__note__`) to ease debugging.
            * The ``where`` queries are now normalized for easier comparison.
            * The assertion output now shows the executed queries on the
              left-hand side and the expected queries on the right-hand side,
              like most other assertion functions.
            * The number of expected and executed queries no longer need to
              be exact in order to see results.

        Version Added:
            3.0

        Args:
            queries (list of ExpectedQuery):
                The list of query dictionaries to compare executed queries
                against.

            num_statements (int, optional):
                The numbre of SQL statements executed.

                This defaults to the length of ``queries``, but callers may
                need to provide an explicit number, as some operations may add
                additional database-specific statements (such as
                transaction-related SQL) that won't be covered in ``queries``.

            with_tracebacks (bool, optional):
                If enabled, tracebacks for queries will be included in
                results.

                Version Added:
                    3.4

            tracebacks_size (int, optional):
                The size of any tracebacks, in number of lines.

                The default is 15.

                Version Added:
                    3.4

            check_join_types (bool, optional):
                Whether to check join types.

                If enabled, table join types (``join_types`` on queries) will
                be checked. This is currently disabled by default, in order
                to avoid breaking tests, but will be enabled by default in
                Djblets 5.

                Version Added:
                    3.4

            check_subqueries (bool, optional):
                Whether to check subqueries.

                If enabled, ``inner_query`` on queries with subqueries will
                be checked. This is currently disabled by default, in order
                to avoid breaking tests, but will be enabled by default in
                Djblets 5.

                Version Added:
                    3.4

        Raises:
            AssertionError:
                The parameters passed, or the queries compared, failed
                expectations.
        """
        def _serialize_mismatched_attrs(
            failures: List[QueryMismatchedAttr],
            *,
            indent: str,
        ) -> List[str]:
            error_lines: List[str] = []

            for mismatched_attr in sorted(failures,
                                          key=lambda attr: attr['name']):
                name = mismatched_attr['name']

                executed_value = mismatched_attr.get('executed_value')
                expected_value = mismatched_attr.get('expected_value')

                assert executed_value is not None
                assert expected_value is not None

                # If we're formatting multi-line output, make sure to
                # indent it properly.
                if '\n' in executed_value or '\n' in expected_value:
                    executed_value = f'\n%s\n{indent}' % '\n'.join(
                        f'{indent}  {line}'
                        for line in executed_value.splitlines()
                    )
                    expected_value = '\n%s' % '\n'.join(
                        f'{indent}  {line}'
                        for line in expected_value.splitlines()
                    )

                error_lines.append(
                    f'{indent}{name}: '
                    f'{executed_value} != {expected_value}')

            return error_lines

        def _serialize_results(
            results: CompareQueriesContext,
            *,
            indent: str = '',
        ) -> List[str]:
            mismatches = results['query_mismatches']
            num_queries = results['num_expected_queries']
            num_executed_queries = results['num_executed_queries']
            inner_indent = f'{indent}  '
            found_subqueries: bool = False

            # Check if we found any failures, and include them in an assertion.
            error_lines: List[str] = []

            if num_queries != num_executed_queries:
                error_lines += [
                    f'{indent}Expected {num_queries} queries, but got '
                    f'{num_executed_queries}',

                    '',
                ]

            if results['has_mismatches']:
                num_mismatches = len(mismatches)

                if num_mismatches == 1:
                    error_lines.append(
                        f'{indent}1 query failed to meet expectations.')
                else:
                    error_lines.append(
                        f'{indent}{num_mismatches} queries failed to meet '
                        f'expectations.'
                    )

                for mismatch_info in mismatches:
                    mismatched_attrs = mismatch_info['mismatched_attrs']
                    subqueries = mismatch_info['subqueries']
                    has_mismatched_subqueries = (check_subqueries and
                                                 subqueries is not None and
                                                 subqueries['has_mismatches'])

                    if subqueries:
                        found_subqueries = True

                    if mismatched_attrs or has_mismatched_subqueries:
                        i = mismatch_info['index']
                        note = mismatch_info['note']
                        traceback = mismatch_info.get('traceback')
                        query_sql = mismatch_info.get('query_sql') or []

                        if note:
                            title = f'{indent}Query {i + 1} ({note}):'
                        else:
                            title = f'{indent}Query {i + 1}:'

                        error_lines += [
                            '',
                            title,
                        ] + _serialize_mismatched_attrs(mismatched_attrs,
                                                        indent=inner_indent)

                        if has_mismatched_subqueries:
                            assert subqueries is not None

                            if mismatched_attrs:
                                error_lines.append('')

                            error_lines += [
                                f'{inner_indent}Subqueries:',
                            ] + _serialize_results(
                                subqueries,
                                indent=f'{inner_indent}  ')

                        if query_sql:
                            error_lines += [
                                f'{indent}  SQL: {_sql}'
                                for _sql in query_sql
                            ]

                        if with_tracebacks and traceback:
                            traceback_str = \
                                ''.join(traceback[-traceback_size:])
                            error_lines.append(
                                f'{indent}Trace: {traceback_str}')

            return error_lines

        # Run the query comparisons.
        with compare_queries(_check_join_types=bool(check_join_types),
                             _check_subqueries=bool(check_subqueries),
                             queries=queries) as results:
            yield

        if results['has_mismatches']:
            self.fail('\n'.join(_serialize_results(results)))


class TestModelsLoaderMixin(object):
    """Allows unit test modules to provide models to test against.

    This allows a unit test file to provide models that will be synced to the
    database and flushed after tests. These can be tested against in any unit
    tests.

    Typically, Django requires any test directories to be pre-added to
    INSTALLED_APPS in order for models to be created in the test database.

    This mixin works around this by dynamically adding the module to
    INSTALLED_APPS and forcing the database to be synced. It also will
    generate a fake 'models' module to satisfy Django's requirement, if one
    doesn't already exist.

    By default, this will assume that the test class's module is the one that
    should be added to INSTALLED_APPS. This can be changed by overriding
    :py:attr:`tests_app`.
    """
    tests_app = None

    @classmethod
    def setUpClass(cls):
        cls._tests_loader_models_mod = None

        if not cls.tests_app:
            cls.tests_app = cls.__module__

        models_mod_name = '%s.models' % cls.tests_app

        try:
            models_mod = import_module(models_mod_name)
        except ImportError:
            # Set up a 'models' module, containing any models local to the
            # module that this TestCase is in.

            # It's not enough to simply create a module type. We need to
            # create a basic spec, and then we need to have the module
            # system create a module from it. There's a handy public
            # function to do this on Python 3.5, but Python 3.4 lacks a
            # public function. Fortunately, it's easy to call a private
            # one.
            spec = ModuleSpec(name=models_mod_name,
                              loader=None)

            models_mod = module_from_spec(spec)
            assert models_mod

            # Transfer all the models over into this new module.
            module_name = cls.__module__
            test_module = sys.modules[module_name]

            for key, value in test_module.__dict__.items():
                if (inspect.isclass(value) and
                    issubclass(value, Model) and
                    value.__module__ == module_name):
                    models_mod.__dict__[key] = value

        cls._tests_loader_models_mod = models_mod

        if models_mod:
            sys.modules[models_mod.__name__] = models_mod

        cls._models_loader_old_settings = settings.INSTALLED_APPS
        settings.INSTALLED_APPS = list(settings.INSTALLED_APPS) + [
            cls.tests_app,
        ]

        # If Django Evolution is being used, we'll want to clear out any
        # recorded schema information so that it can be generated from
        # scratch when we set up the database.
        if (Evolution is not None and
            Version is not None and
            getattr(settings, 'DJANGO_EVOLUTION_ENABLED', True)):
            Version.objects.all().delete()
            Evolution.objects.all().delete()

        # Push the new set of installed apps, and begin registering
        # each of the models associated with the tests.
        apps.set_installed_apps(settings.INSTALLED_APPS)
        app_config = apps.get_containing_app_config(cls.tests_app)

        if models_mod and app_config:
            app_label = app_config.label

            for key, value in models_mod.__dict__.items():
                if inspect.isclass(value) and issubclass(value, Model):
                    # The model was likely registered under another app,
                    # so we need to remove the old one and add the new
                    # one.
                    if value._meta.model_name:
                        try:
                            del apps.all_models[value._meta.app_label][
                                value._meta.model_name]
                        except KeyError:
                            pass

                    value._meta.app_label = app_label
                    apps.register_model(app_label, value)

        call_command('migrate', run_syncdb=True, verbosity=0,
                     interactive=False)

        super(TestModelsLoaderMixin, cls).setUpClass()

    @classmethod
    def tearDownClass(cls):
        super(TestModelsLoaderMixin, cls).tearDownClass()

        call_command('flush', verbosity=0, interactive=False)

        settings.INSTALLED_APPS = cls._models_loader_old_settings

        # If we added a fake 'models' module to sys.modules, remove it.
        models_mod = cls._tests_loader_models_mod

        if models_mod:
            try:
                del sys.modules[models_mod.__name__]
            except KeyError:
                pass

        assert cls.tests_app

        apps.unset_installed_apps()
        apps.all_models[cls.tests_app].clear()
        apps.populate(settings.INSTALLED_APPS)

        # Set this free so the garbage collector can eat it.
        cls._tests_loader_models_mod = None


class FixturesCompilerMixin(object):
    """Compiles and efficiently loads fixtures into a test suite.

    Unlike Django's standard fixture support, this doesn't re-discover
    and re-deserialize the referenced fixtures every time they're needed.
    Instead, it precompiles the fixtures the first time they're found and
    reuses their objects for future tests.

    However, also unlike Django's, this does not accept compressed or
    non-JSON fixtures.
    """

    _precompiled_fixtures = {}
    _fixture_dirs = []

    def _fixture_setup(self):
        """Set up fixtures for unit tests."""
        # Temporarily hide the fixtures, so that the parent class won't
        # do anything with them.
        self._hide_fixtures = True
        super(FixturesCompilerMixin, self)._fixture_setup()
        self._hide_fixtures = False

        if getattr(self, 'multi_db', False):
            databases = connections
        else:
            databases = [DEFAULT_DB_ALIAS]

        if hasattr(self, 'fixtures'):
            for db in databases:
                self.load_fixtures(self.fixtures, db=db)

    def load_fixtures(self, fixtures, db=DEFAULT_DB_ALIAS):
        """Load fixtures for the current test.

        This is called for every fixture in the test case's ``fixtures``
        list. It can also be called by an individual test to add additional
        fixtures on top of that.

        Args:
            fixtures (list of unicode):
                The list of fixtures to load.

            db (unicode):
                The database name to load fixture data on.
        """
        if not fixtures:
            return

        if db not in self._precompiled_fixtures:
            self._precompiled_fixtures[db] = {}

        for fixture in fixtures:
            if fixture not in self._precompiled_fixtures[db]:
                self._precompile_fixture(fixture, db)

        self._load_fixtures(fixtures, db)

    def _precompile_fixture(self, fixture, db):
        """Precompile a fixture.

        The fixture is loaded and deserialized, and the resulting objects
        are stored for future use.

        Args:
            fixture (unicode):
                The name of the fixture.

            db (unicode):
                The database name to load fixture data on.
        """
        assert db in self._precompiled_fixtures
        assert fixture not in self._precompiled_fixtures[db]

        fixture_path = None

        for fixture_dir in self._get_fixture_dirs():
            fixture_path = os.path.join(fixture_dir, fixture + '.json')

            if os.path.exists(fixture_path):
                break

        try:
            if not fixture_path:
                raise IOError('Fixture path not found')

            with open(fixture_path, 'r') as fp:
                self._precompiled_fixtures[db][fixture] = [
                    obj
                    for obj in serializers.deserialize('json', fp, using=db)
                    if ((hasattr(router, 'allow_syncdb') and
                         router.allow_syncdb(db, obj.object.__class__)) or
                        (hasattr(router, 'allow_migrate_model') and
                         router.allow_migrate_model(db, obj.object)))
                ]
        except IOError as e:
            sys.stderr.write('Unable to load fixture %s: %s\n' % (fixture, e))

    def _get_fixture_dirs(self):
        """Return the list of fixture directories.

        This is computed only once and cached.

        Returns:
            The list of fixture directories.
        """
        if not self._fixture_dirs:
            app_module_paths = []

            for app in self._get_app_model_modules():
                if hasattr(app, '__path__'):
                    # It's a 'models/' subpackage.
                    for path in app.__path__:
                        app_module_paths.append(path)
                else:
                    # It's a models.py module
                    app_module_paths.append(app.__file__)

            all_fixture_dirs = [
                os.path.join(os.path.dirname(path), 'fixtures')
                for path in app_module_paths
            ]

            self._fixture_dirs = [
                fixture_dir
                for fixture_dir in all_fixture_dirs
                if os.path.exists(fixture_dir)
            ]

        return self._fixture_dirs

    def _load_fixtures(self, fixtures, db):
        """Load precompiled fixtures.

        Each precompiled fixture is loaded and then used to populate the
        database.

        Args:
            fixtures (list of unicode):
                The list of fixtures to load.

            db (unicode):
                The database name to load fixture data on.
        """
        table_names = set()
        connection = connections[db]

        with connection.constraint_checks_disabled():
            for fixture in fixtures:
                assert db in self._precompiled_fixtures
                assert fixture in self._precompiled_fixtures[db]
                deserialized_objs = self._precompiled_fixtures[db][fixture]

                to_save = OrderedDict()
                to_save_m2m = []

                # Start off by going through the list of deserialized object
                # information from the fixtures and group them into categories
                # that we'll further filter down later.
                for deserialized_obj in deserialized_objs:
                    obj = deserialized_obj.object

                    table_names.add(obj._meta.db_table)
                    to_save.setdefault(type(obj), []).append(obj)

                    if deserialized_obj.m2m_data:
                        to_save_m2m.append((obj, deserialized_obj.m2m_data))

                # Now we'll break this down into objects we can batch-create
                # and ones we have to update.
                #
                # The database may already have entries for some of these
                # models, particularly if they're being shared across multiple
                # tests for the same suite. In that case, when doing a save()
                # on an object, Django will perform a update-or-create
                # operation, involving an UPDATE statement followed by an
                # INSERT If the UPDATE returned 0 rows.
                #
                # That's wasteful. Instead, we're going to just delete
                # everything matching the deserialized objects' IDs and then
                # bulk-create new entries.
                try:
                    for model_cls, objs in to_save.items():
                        obj_ids = [
                            _obj.pk
                            for _obj in objs
                            if _obj.pk
                        ]

                        if obj_ids:
                            model_cls.objects.filter(pk__in=obj_ids).delete()

                        model_cls.objects.using(db).bulk_create(objs)
                except (DatabaseError, IntegrityError) as e:
                    meta = model_cls._meta
                    sys.stderr.write(
                        'Could not load %s.%s from fixture "%s": %s\n'
                        % (meta.app_label,
                           meta.object_name,
                           fixture,
                           e))
                    raise

                # Now save any Many-to-Many field relations.
                #
                # Note that we're not assigning the values to the
                # ManyToManyField attribute. Instead, we're getting the add()
                # method and calling that. We can trust that the relations are
                # empty (since we've newly-created the object), so we don't
                # need to clear the old list first.
                for obj, m2m_data in to_save_m2m:
                    try:
                        for m2m_attr, m2m_objs in m2m_data.items():
                            getattr(obj, m2m_attr).add(*m2m_objs)
                    except (DatabaseError, IntegrityError) as e:
                        meta = obj._meta
                        sys.stderr.write(
                            'Could not load Many-to-Many entries for %s.%s '
                            '(pk=%s) in fixture "%s": %s\n'
                            % (meta.app_label,
                               meta.object_name,
                               obj.pk,
                               fixture,
                               e))
                        raise

        # We disabled constraints above, so check now.
        connection.check_constraints(table_names=sorted(table_names))

    def _get_app_model_modules(self):
        """Return the entire list of registered Django app models modules.

        This is an internal utility function that's used to provide
        compatibility with all supported versions of Django.

        Returns:
            list:
            The list of Django applications.
        """
        return [
            app.models_module
            for app in apps.get_app_configs()
            if app.models_module
        ]

    def __getattribute__(self, name):
        if name == 'fixtures' and self.__dict__.get('_hide_fixtures'):
            raise AttributeError

        return super(FixturesCompilerMixin, self).__getattribute__(name)


class TagTest(TestCase):
    """Base testing setup for custom template tags"""

    def setUp(self):
        self.parser = StubParser(self.getContentText())

    def getContentText(self):
        return "content"


# The following is all based on the code at
# http://trac.getwindmill.com/browser/trunk/windmill/authoring/djangotest.py,
# which is based on the changes submitted for Django in ticket 2879
# (http://code.djangoproject.com/ticket/2879)
#
# A lot of this can go away when/if this patch is committed to Django.

# Code from django_live_server_r8458.diff
#     @ http://code.djangoproject.com/ticket/2879#comment:41
# Editing to monkey patch django rather than be in trunk

class StoppableWSGIServer(basehttp.WSGIServer):
    """
    WSGIServer with short timeout, so that server thread can stop this server.
    """
    def server_bind(self):
        """Sets timeout to 1 second."""
        basehttp.WSGIServer.server_bind(self)
        self.socket.settimeout(1)

    def get_request(self):
        """Checks for timeout when getting request."""
        try:
            sock, address = self.socket.accept()
            sock.settimeout(None)
            return (sock, address)
        except socket.timeout:
            raise


class WSGIRequestHandler(basehttp.WSGIRequestHandler):
    """A custom WSGIRequestHandler that logs all output to stdout.

    Normally, WSGIRequestHandler will color-code messages and log them
    to stderr. It also filters out admin and favicon.ico requests. We don't
    need any of this, and certainly don't want it in stderr, as we'd like
    to only show it on failure.
    """
    def log_message(self, format, *args):
        print(format % args)


class TestServerThread(threading.Thread):
    """Thread for running a http server while tests are running."""

    def __init__(self, address, port):
        self.address = address
        self.port = port
        self._stopevent = threading.Event()
        self.started = threading.Event()
        self.error = None
        super(TestServerThread, self).__init__()

    def run(self):
        """
        Sets up test server and database and loops over handling http requests.
        """
        try:
            handler = basehttp.AdminMediaHandler(WSGIHandler())
            server_address = (self.address, self.port)
            httpd = StoppableWSGIServer(server_address,
                                        WSGIRequestHandler)
            httpd.set_app(handler)
            self.started.set()
        except basehttp.WSGIServerException as e:
            self.error = e
            self.started.set()
            return

        # Must do database stuff in this new thread if database in memory.
        from django.conf import settings

        if hasattr(settings, 'DATABASES'):
            db_engine = settings.DATABASES['default']['ENGINE']
            test_db_name = settings.DATABASES['default']['TEST_NAME']
        else:
            db_engine = settings.DATABASE_ENGINE
            test_db_name = settings.TEST_DATABASE_NAME

        if (db_engine.endswith('sqlite3') and
            (not test_db_name or test_db_name == ':memory:')):
            # Import the fixture data into the test database.
            if hasattr(self, 'fixtures'):
                # We have to use this slightly awkward syntax due to the fact
                # that we're using *args and **kwargs together.
                testcases.call_command('loaddata', verbosity=0, *self.fixtures)

        # Loop until we get a stop event.
        while not self._stopevent.isSet():
            httpd.handle_request()

    def join(self, timeout=None):
        """Stop the thread and wait for it to finish."""
        self._stopevent.set()
        threading.Thread.join(self, timeout)



"""Unit tests for djblets.db.query_catcher.

Version Added:
    3.4
"""

from __future__ import annotations

import re
from typing import Any, List, Optional, TYPE_CHECKING, Type

from django.db import models
from django.db.models import Exists, OuterRef, Q, QuerySet, Subquery, Sum
from django.db.models.sql.subqueries import AggregateQuery

from djblets.db.query_catcher import (ExecutedQueryType,
                                      catch_queries)
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin

if TYPE_CHECKING:
    from djblets.db.query_catcher import (CatchQueriesContext,
                                          ExecutedQueryInfo,
                                          ExecutedSubQueryInfo)


class CapturedQueriesTestModel(models.Model):
    name = models.CharField(max_length=10)
    flag = models.BooleanField(default=True)


class CapturedQueriesRelTestModel(models.Model):
    # Set up a relation, but don't do anything if the test model deletes.
    # We don't want this interfering with our query tests.
    test = models.ForeignKey(CapturedQueriesTestModel,
                             on_delete=models.DO_NOTHING)


class CaptureQueriesTests(TestModelsLoaderMixin, TestCase):
    """Unit tests for djblets.db.query_catcher.catch_queries.

    Version Added:
        3.4
    """

    tests_app = 'djblets.db.tests'

    maxDiff = 10_000

    _extra_ws_re = re.compile(r'\s{2,}')

    def test_with_select(self) -> None:
        """Testing capture_queries with SELECT"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        with catch_queries() as ctx:
            objs = list(
                CapturedQueriesTestModel.objects
                .filter(Q(name__startswith='test') &
                        Q(flag=True))
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            sql=[
                'SELECT'
                ' "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag" '
                ' FROM "djblets_db_capturedqueriestestmodel"'
                ' WHERE'
                '  ("djblets_db_capturedqueriestestmodel"."name"'
                '   LIKE test% ESCAPE \'\\\''
                '   AND "djblets_db_capturedqueriestestmodel"."flag")',
            ],
            q=(
                Q(Q(name__startswith='test') &
                  Q(flag=True))
            ))

        self.assertEqual(len(objs), 2)
        self.assertEqual(objs[0].name, 'test1')
        self.assertEqual(objs[1].name, 'test2')

    def test_with_insert(self) -> None:
        """Testing capture_queries with SELECT"""
        with catch_queries() as ctx:
            CapturedQueriesTestModel.objects.bulk_create([
                CapturedQueriesTestModel(name='test1'),
                CapturedQueriesTestModel(name='test2'),
            ])

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            query_type=ExecutedQueryType.INSERT,
            sql=[
                'INSERT INTO "djblets_db_capturedqueriestestmodel"'
                ' ("name", "flag")'
                ' VALUES (test1, True), (test2, True)',
            ])

    def test_with_delete(self) -> None:
        """Testing capture_queries with DELETE"""
        obj = CapturedQueriesTestModel.objects.create(name='test')
        obj_id = obj.pk

        with catch_queries() as ctx:
            obj.delete()

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            query_type=ExecutedQueryType.DELETE,
            sql=[
                'DELETE FROM "djblets_db_capturedqueriestestmodel"'
                ' WHERE "djblets_db_capturedqueriestestmodel"."id" IN (1)',
            ],
            q=Q(id__in=[obj_id]))

    def test_with_update(self) -> None:
        """Testing capture_queries with UPDATE"""
        obj = CapturedQueriesTestModel.objects.create(name='test')

        with catch_queries() as ctx:
            obj.flag = False
            obj.save(update_fields=('flag',))

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            query_type=ExecutedQueryType.UPDATE,
            sql=[
                'UPDATE "djblets_db_capturedqueriestestmodel"'
                ' SET "flag" = False'
                ' WHERE "djblets_db_capturedqueriestestmodel"."id" = 1',
            ],
            q=Q(pk=obj.pk))

    def test_with_multiple(self) -> None:
        """Testing capture_queries with multiple queries"""
        with catch_queries() as ctx:
            obj = CapturedQueriesTestModel.objects.create(name='test')
            obj.flag = False
            obj.save(update_fields=('flag',))

            list(CapturedQueriesTestModel.objects.all())

            obj.delete()

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 4)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            query_type=ExecutedQueryType.INSERT,
            sql=[
                'INSERT INTO "djblets_db_capturedqueriestestmodel"'
                ' ("name", "flag")'
                ' VALUES (test, True)',
            ])

        self._check_query(
            executed_queries[1],
            ctx=ctx,
            query_type=ExecutedQueryType.UPDATE,
            sql=[
                'UPDATE "djblets_db_capturedqueriestestmodel"'
                ' SET "flag" = False'
                ' WHERE "djblets_db_capturedqueriestestmodel"."id" = 1',
            ],
            q=Q(pk=1))

        self._check_query(
            executed_queries[2],
            ctx=ctx,
            query_type=ExecutedQueryType.SELECT,
            sql=[
                'SELECT "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag"'
                ' FROM "djblets_db_capturedqueriestestmodel"',
            ])

        self._check_query(
            executed_queries[3],
            ctx=ctx,
            query_type=ExecutedQueryType.DELETE,
            sql=[
                'DELETE FROM "djblets_db_capturedqueriestestmodel"'
                ' WHERE "djblets_db_capturedqueriestestmodel"."id" IN (1)',
            ],
            q=Q(id__in=[1]))

    def test_with_annotation_subquery(self) -> None:
        """Testing capture_queries with annotation and Subquery"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[0]),
        ])

        with catch_queries() as ctx:
            objs = list(
                CapturedQueriesTestModel.objects
                .filter(Q(name__startswith='test'))
                .annotate(sub=Subquery(
                    CapturedQueriesRelTestModel.objects
                    .filter(test=OuterRef('pk'))
                    .values('pk')
                ))
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT'
                ' "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag", '
                ' (SELECT U0."id"'
                '   FROM "djblets_db_capturedqueriesreltestmodel" U0'
                '   WHERE'
                '    U0."test_id" ='
                '    ("djblets_db_capturedqueriestestmodel"."id")) AS "sub"'
                ' FROM "djblets_db_capturedqueriestestmodel" '
                ' WHERE'
                '  "djblets_db_capturedqueriestestmodel"."name"'
                '   LIKE test% ESCAPE \'\\\''
            ],

            # NOTE: Subquery() cannot be compared through equality checks.
            can_compare_q=False)

        # Check the subqueries.
        self._check_subquery(
            executed_queries[0]['subqueries'][0],
            ctx=ctx,
            subquery_class=Subquery,
            q=Q(test=OuterRef('pk')))

        # Check the fetched objects from the query.
        self.assertEqual(len(objs), 2)
        self.assertEqual(objs[0].name, 'test1')
        self.assertEqual(objs[1].name, 'test2')

    def test_with_annotation_exists(self) -> None:
        """Testing capture_queries with annotation and Exists"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[0]),
        ])

        with catch_queries() as ctx:
            objs = list(
                CapturedQueriesTestModel.objects
                .filter(Q(name__startswith='test'))
                .annotate(sub=Exists(
                    CapturedQueriesRelTestModel.objects
                    .filter(test=OuterRef('pk'))
                ))
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT'
                ' "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag", '
                ' EXISTS(SELECT 1 AS "a"'
                '  FROM "djblets_db_capturedqueriesreltestmodel" U0'
                '  WHERE'
                '   U0."test_id" ='
                '   ("djblets_db_capturedqueriestestmodel"."id")'
                '  LIMIT 1) AS "sub"'
                ' FROM "djblets_db_capturedqueriestestmodel" '
                ' WHERE'
                '  "djblets_db_capturedqueriestestmodel"."name"'
                '   LIKE test% ESCAPE \'\\\''
            ],

            # NOTE: Exists() cannot be compared through equality checks.
            can_compare_q=False)

        # Check the subqueries.
        self._check_subquery(
            executed_queries[0]['subqueries'][0],
            ctx=ctx,
            subquery_class=Exists,
            q=Q(test=OuterRef('pk')))

        # Check the fetched objects from the query.
        self.assertEqual(len(objs), 2)
        self.assertEqual(objs[0].name, 'test1')
        self.assertEqual(objs[1].name, 'test2')

    def test_with_aggregate_subquery(self) -> None:
        """Testing capture_queries with aggregates and Subquery"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[1]),
            CapturedQueriesRelTestModel(test=test_models[1]),
        ])

        with catch_queries() as ctx:
            # This query is a particular amount of nonsense. Don't worry
            # about it.
            list(
                CapturedQueriesTestModel.objects
                .filter(
                    Q(name__startswith='test') &
                    Q(pk__gt=Subquery(
                        CapturedQueriesRelTestModel.objects
                        .filter(test=OuterRef('pk'))
                        .annotate(some_value=Sum('pk') + 1)
                        .values('some_value')
                    ))
                )
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT'
                ' "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag"'
                ' FROM "djblets_db_capturedqueriestestmodel" '
                ' WHERE'
                '  ("djblets_db_capturedqueriestestmodel"."name"'
                '    LIKE test% ESCAPE \'\\\' AND'
                '   "djblets_db_capturedqueriestestmodel"."id" > '
                '    (SELECT (SUM(U0."id") + 1) AS "some_value"'
                '      FROM "djblets_db_capturedqueriesreltestmodel" U0'
                '      WHERE U0."test_id" ='
                '       ("djblets_db_capturedqueriestestmodel"."id")'
                '      GROUP BY U0."id", U0."test_id"))'
            ],

            # NOTE: Subquery() cannot be compared through equality checks.
            can_compare_q=False)

        # Check the subqueries.
        self._check_subquery(
            executed_queries[0]['subqueries'][0],
            ctx=ctx,
            subquery_class=Subquery,
            q=Q(test=OuterRef('pk')))

    def test_with_filter_subquery_exists(self) -> None:
        """Testing capture_queries with filtering using Exists subqueries"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[0]),
        ])

        with catch_queries() as ctx:
            objs = list(
                CapturedQueriesTestModel.objects
                .filter(
                    Q(name__startswith='test') &
                    Exists(
                        CapturedQueriesRelTestModel.objects
                        .filter(test=OuterRef('pk'))
                    )
                )
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT'
                ' "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag"'
                ' FROM "djblets_db_capturedqueriestestmodel" '
                ' WHERE'
                '  ("djblets_db_capturedqueriestestmodel"."name"'
                '    LIKE test% ESCAPE \'\\\' AND'
                '   EXISTS(SELECT 1 AS "a"'
                '    FROM "djblets_db_capturedqueriesreltestmodel" U0'
                '    WHERE U0."test_id" ='
                '     ("djblets_db_capturedqueriestestmodel"."id")'
                '    LIMIT 1))',
            ],

            # NOTE: Exists() cannot be compared through equality checks.
            can_compare_q=False)

        # Check the subqueries.
        self._check_subquery(
            executed_queries[0]['subqueries'][0],
            ctx=ctx,
            subquery_class=Exists,
            q=Q(test=OuterRef('pk')))

        # Check the fetched objects from the query.
        self.assertEqual(len(objs), 1)
        self.assertEqual(objs[0].name, 'test1')

    def test_with_filter_subquery_q(self) -> None:
        """Testing capture_queries with filtering using Q subqueries"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[0]),
        ])

        with catch_queries() as ctx:
            objs = list(
                CapturedQueriesTestModel.objects
                .filter(
                    Q(name__startswith='test') &
                    Q(pk__in=(
                        CapturedQueriesRelTestModel.objects
                        .values_list('pk', flat=True)
                    ))
                )
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT'
                ' "djblets_db_capturedqueriestestmodel"."id",'
                ' "djblets_db_capturedqueriestestmodel"."name",'
                ' "djblets_db_capturedqueriestestmodel"."flag"'
                ' FROM "djblets_db_capturedqueriestestmodel" '
                ' WHERE'
                '  ("djblets_db_capturedqueriestestmodel"."name"'
                '    LIKE test% ESCAPE \'\\\' AND'
                '   "djblets_db_capturedqueriestestmodel"."id" IN'
                '    (SELECT U0."id" FROM'
                '      "djblets_db_capturedqueriesreltestmodel" U0))',
            ],

            # NOTE: QuerySet() cannot be compared consistently through
            #       equality checks.
            can_compare_q=False)

        # Check the subqueries.
        self._check_subquery(
            executed_queries[0]['subqueries'][0],
            ctx=ctx,
            subquery_class=QuerySet)

        # Check the fetched objects from the query.
        self.assertEqual(len(objs), 2)
        self.assertEqual(objs[0].name, 'test1')
        self.assertEqual(objs[1].name, 'test2')

    def test_with_distinct_count(self) -> None:
        """Testing capture_queries with distinct() and count()"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[0]),
        ])

        with catch_queries() as ctx:
            count = (
                CapturedQueriesTestModel.objects
                .filter(name__startswith='test')
                .distinct()
                .count()
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT COUNT(*)'
                ' FROM'
                ' (SELECT DISTINCT'
                '   "djblets_db_capturedqueriestestmodel"."id" AS "col1",'
                '   "djblets_db_capturedqueriestestmodel"."name" AS "col2",'
                '   "djblets_db_capturedqueriestestmodel"."flag" AS "col3"'
                '  FROM "djblets_db_capturedqueriestestmodel"'
                '  WHERE "djblets_db_capturedqueriestestmodel"."name"'
                '   LIKE test% ESCAPE \'\\\') subquery'
            ])

        # Check the subqueries.
        self._check_subquery(
            executed_queries[0]['subqueries'][0],
            ctx=ctx,
            subquery_class=AggregateQuery,
            q=Q(name__startswith='test'))

        # Check the fetched count.
        self.assertEqual(count, 2)

    def test_with_subqueries_complex(self) -> None:
        """Testing capture_queries with complex subqueries"""
        CapturedQueriesTestModel.objects.bulk_create([
            CapturedQueriesTestModel(name='test1'),
            CapturedQueriesTestModel(name='test2'),
        ])

        # Re-fetch to guarantee IDs.
        test_models = list(CapturedQueriesTestModel.objects.all())

        CapturedQueriesRelTestModel.objects.bulk_create([
            CapturedQueriesRelTestModel(test=test_models[0]),
            CapturedQueriesRelTestModel(test=test_models[0]),
        ])

        with catch_queries() as ctx:
            count = (
                CapturedQueriesTestModel.objects
                .filter(
                    Q(name__startswith='test') &
                    Exists(
                        CapturedQueriesRelTestModel.objects
                        .filter(
                            Q(test=OuterRef('pk')) &
                            ~Exists(
                                CapturedQueriesRelTestModel.objects
                                .filter(test=2)
                            )
                        )
                    )
                )
                .annotate(
                    sub1=Subquery(
                        CapturedQueriesRelTestModel.objects
                        .filter(
                            Q(test=OuterRef('pk')) &
                            Q(test__gt=Subquery(
                                CapturedQueriesRelTestModel.objects
                                .filter(test=OuterRef('pk'))
                                .annotate(some_value=Sum('pk') + 1)
                                .values('some_value')
                            ))
                        )
                        .values('pk')
                    ),
                )
                .distinct()
                .count()
            )

        executed_queries = ctx.executed_queries
        self.assertEqual(len(executed_queries), 1)

        self._check_query(
            executed_queries[0],
            ctx=ctx,
            num_subqueries=1,
            sql=[
                'SELECT COUNT(*) FROM'
                ' (SELECT DISTINCT'
                '  "djblets_db_capturedqueriestestmodel"."id" AS "col1",'
                '  "djblets_db_capturedqueriestestmodel"."name" AS "col2",'
                '  "djblets_db_capturedqueriestestmodel"."flag" AS "col3",'
                '  (SELECT V0."id"'
                '    FROM "djblets_db_capturedqueriesreltestmodel" V0'
                '    WHERE (V0."test_id" ='
                '     ("djblets_db_capturedqueriestestmodel"."id") AND'
                '     V0."test_id" >'
                '     (SELECT (SUM(U0."id") + 1) AS "some_value"'
                '       FROM "djblets_db_capturedqueriesreltestmodel" U0'
                '       WHERE U0."test_id" = (V0."id")'
                '       GROUP BY U0."id", U0."test_id"))) AS "sub1"'
                '  FROM "djblets_db_capturedqueriestestmodel"'
                '  WHERE'
                '   ("djblets_db_capturedqueriestestmodel"."name"'
                '     LIKE test% ESCAPE \'\\\' AND'
                '    EXISTS(SELECT 1 AS "a"'
                '     FROM "djblets_db_capturedqueriesreltestmodel" V0'
                '     WHERE'
                '      (V0."test_id" ='
                '       ("djblets_db_capturedqueriestestmodel"."id")'
                '       AND NOT EXISTS(SELECT 1 AS "a"'
                '        FROM "djblets_db_capturedqueriesreltestmodel" U0'
                '        WHERE U0."test_id" = 2'
                '        LIMIT 1))'
                '     LIMIT 1))) subquery',
            ])

        # Check the distinct subquery.
        distinct_subquery = executed_queries[0]['subqueries'][0]

        self._check_subquery(distinct_subquery,
                             ctx=ctx,
                             subquery_class=AggregateQuery,
                             num_subqueries=2,
                             can_compare_q=False)

        subqueries = distinct_subquery['subqueries']

        # Check annotate subquery 1.
        self._check_subquery(subqueries[0],
                             ctx=ctx,
                             subquery_class=Subquery,
                             num_subqueries=1,
                             can_compare_q=False)  # Subquery() again.
        self._check_subquery(subqueries[0]['subqueries'][0],
                             ctx=ctx,
                             subquery_class=Subquery,
                             can_compare_q=False)  # Subquery() again.

        # Check filter subquery 1.
        self._check_subquery(subqueries[1],
                             ctx=ctx,
                             subquery_class=Exists,
                             num_subqueries=1,
                             can_compare_q=False)   # Exists() again.
        self._check_subquery(subqueries[1]['subqueries'][0],
                             subquery_class=Exists,
                             ctx=ctx,
                             q=Q(test=2))

        # Check the fetched count.
        self.assertEqual(count, 1)

    def _check_query(
        self,
        executed_query: ExecutedQueryInfo,
        *,
        ctx: CatchQueriesContext,
        sql: List[str],
        num_subqueries: int = 0,
        query_type: ExecutedQueryType = ExecutedQueryType.SELECT,
        q: Optional[Q] = None,
        can_compare_q: bool = True,
    ) -> None:
        """Check the recorded information on a query.

        This will perform stanadrd checks on query state, and compare the
        provided attributes.

        Args:
            executed_query (dict):
                Information on the query that was executed.

            ctx (dict):
                The context information from catching the queries.

            sql (list of str):
                The expected list of SQL statements for the query.

            num_subqueries (int, optional):
                The expected number of subqueries within this query.

            query_type (djblets.db.query_catcher.ExecutedQueryType, optional):
                The expected query type.

                This defaults to ``SELECT`` queries.

            q (django.db.models.Q, optional):
                The Q object for the query, if any.

                This can be omitted if passing ``can_compare_q=False`` or
                for queries that lack filtering.

            can_compare_q (bool, optional):
                Whether ``q`` can be compared.

                If ``False``, this will merely check for the presence of a
                recorded Q object.

        Raises:
            AssertionError:
                One or more checks failed.
        """
        self.assertEqual(executed_query['result_type'], 'query')
        self.assertEqual(executed_query['type'], query_type)
        self._assert_sql(executed_query['sql'], sql)
        self.assertEqual(len(executed_query['subqueries']), num_subqueries)
        self.assertGreaterEqual(len(executed_query['traceback']), 1)

        query = executed_query['query']
        self.assertIsNotNone(query)

        if q is None and can_compare_q:
            self.assertNotIn(query, ctx.queries_to_qs)
        else:
            self.assertIn(query, ctx.queries_to_qs)

            if can_compare_q:
                self.assertEqual(ctx.queries_to_qs[query], q)

    def _check_subquery(
        self,
        executed_subquery: ExecutedSubQueryInfo,
        *,
        ctx: CatchQueriesContext,
        subquery_class: Type[Any],
        num_subqueries: int = 0,
        q: Optional[Q] = None,
        can_compare_q: bool = True,
    ) -> None:
        """Check the recorded information on a subquery.

        This will perform stanadrd checks on subquery state, and compare the
        provided attributes.

        Args:
            executed_subquery (dict):
                Information on the subquery that was executed.

            ctx (dict):
                The context information from catching the queries.

            subquery_class (type):
                The expected class for the subquery.

            num_subqueries (int, optional):
                The expected number of subqueries within this subquery.

            q (django.db.models.Q, optional):
                The Q object for the subquery, if any.

                This can be omitted if passing ``can_compare_q=False`` or
                for queries that lack filtering.

            can_compare_q (bool, optional):
                Whether ``q`` can be compared.

                If ``False``, this will merely check for the presence of a
                recorded Q object.

        Raises:
            AssertionError:
                One or more checks failed.
        """
        self.assertEqual(executed_subquery['result_type'], 'subquery')
        self.assertEqual(executed_subquery['type'], ExecutedQueryType.SELECT)
        self.assertIs(executed_subquery['cls'], subquery_class)
        self.assertEqual(len(executed_subquery['subqueries']), num_subqueries)

        query = executed_subquery['query']
        self.assertIsNotNone(query)

        if q is None and can_compare_q:
            self.assertNotIn(query, ctx.queries_to_qs)
        else:
            self.assertIn(query, ctx.queries_to_qs)

            if can_compare_q:
                self.assertEqual(ctx.queries_to_qs[query], q)

    def _assert_sql(
        self,
        sql1: List[str],
        sql2: List[str],
    ) -> None:
        """Assert that two lists of SQL statements are equal.

        This will normalize the lists, removing any extra whitespace that may
        be added to help format the SQL for editing and display.

        Args:
            sql1 (list of str):
                The first list of SQL statements.

            sql1 (list of str):
                The second list of SQL statements.

        Raises:
            AssertionError:
                The SQL statements are not equal.
        """
        extra_ws_re = self._extra_ws_re
        self.assertEqual(
            '\n'.join(
                extra_ws_re.sub(' ', _line)
                for _line in sql1
            ),
            '\n'.join(
                extra_ws_re.sub(' ', _line)
                for _line in sql2
            ))



"""Common symbols useful for function signatures.

Version Added:
    3.3
"""

from enum import Enum
from typing import TypeVar, Union

from typing_extensions import Final, Literal, TypeAlias


_T = TypeVar('_T')


class UnsetSymbol(Enum):
    """A type indicating an unsettable value.

    This can be useful in functions that take default values to distinguish
    between a value not provided and a false/``None`` value.

    Version Changed:
        4.0:
        This is now an Enum, allowing values to be used in
        :py:class:`typing.Literal`.

    Version Added:
        3.3
    """

    UNSET = '<UNSET>'


#: An instance of a symbol indicating an unset value.
#:
#: Version Added:
#:     3.3
UNSET: Final[Literal[UnsetSymbol.UNSET]] = UnsetSymbol.UNSET


#: A generic type alias for marking a type as unsettable.
#:
#: This allows for usage like:
#:
#: .. code-block:: python
#:
#:    def __init__(
#:        self,
#:        value: Unsettable[str],
#:    ) -> None:
#:        ...
#:
#: Version Added:
#:     4.0
Unsettable: TypeAlias = Union[Literal[UnsetSymbol.UNSET], _T]



"""Tests for the utilities for rate-limiting login attempts."""

import re

import kgb
from django.contrib.auth.models import AnonymousUser, User
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory
from django.test.utils import override_settings

from djblets.auth.ratelimit import (Rate,
                                    _get_time_int,
                                    get_usage_count,
                                    is_ratelimited)
from djblets.testing.testcases import TestCase


class RateLimitTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.auth.ratelimit."""

    def setUp(self):
        super(RateLimitTests, self).setUp()
        self.request_factory = RequestFactory()

        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(1659732041))

    def tearDown(self):
        super(RateLimitTests, self).tearDown()
        cache.clear()

    def test_rate_parsing(self):
        """Testing Rate.parse"""
        test_rates = (
            ('100/s', Rate(100, 1)),
            ('100/10s', Rate(100, 10)),
            ('100/10', Rate(100, 10)),
            ('100/m', Rate(100, 60)),
            ('400/10m', Rate(400, 600)),
            ('1000/h', Rate(1000, 3600)),
            ('800/d', Rate(800, 24 * 60 * 60)),
        )

        for rate_str, rate in test_rates:
            self.assertEqual(rate, Rate.parse(rate_str))

    @override_settings(LOGIN_LIMIT_RATE='7/m')
    def test_unauthenticated_user(self):
        """Testing is_ratelimited with unauthenticated user"""
        request = self.request_factory.get('/')
        request.user = AnonymousUser()

        self.assertFalse(is_ratelimited(request, increment=False))

    @override_settings(LOGIN_LIMIT_RATE='blah')
    def test_invalid_rate_limit(self):
        """Testing is_ratelimited with invalid rate limit parameter"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        with self.assertRaises(ImproperlyConfigured) as context:
            is_ratelimited(request, increment=False)

        self.assertEqual(str(context.exception),
                         'LOGIN_LIMIT_RATE setting could not be parsed.')

    @override_settings(LOGIN_LIMIT_RATE='1/h')
    def test_rate_limit_exceeded(self):
        """Testing is_ratelimited when limit exceeded"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.assertFalse(is_ratelimited(request, increment=True))
        self.assertTrue(is_ratelimited(request, increment=True))

    @override_settings(LOGIN_LIMIT_RATE='1/s')
    def test_get_usage_count_at_1s(self):
        """Testing get_usage_count at 1/s"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 1,
                'limit': 1,
                'time_left': 0,
            })

        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 2,
                'limit': 1,
                'time_left': 0,
            })

    @override_settings(LOGIN_LIMIT_RATE='1/m')
    def test_get_usage_count_at_1m(self):
        """Testing get_usage_count at 1/m"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 1,
                'limit': 1,
                'time_left': 19,
            })

        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 2,
                'limit': 1,
                'time_left': 19,
            })

    @override_settings(LOGIN_LIMIT_RATE='1/h')
    def test_get_usage_count_at_1h(self):
        """Testing get_usage_count at 1/h"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 1,
                'limit': 1,
                'time_left': 1159,
            })

        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 2,
                'limit': 1,
                'time_left': 1159,
            })

    @override_settings(LOGIN_LIMIT_RATE='1/s')
    def test_get_usage_count_with_increment_false(self):
        """Testing get_usage_count with increment=False"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        # First call will always have count=1, no matter what.
        self.assertEqual(
            get_usage_count(request=request,
                            increment=False),

            {
                'count': 1,
                'limit': 1,
                'time_left': 0,
            })

        self.assertEqual(
            get_usage_count(request=request,
                            increment=False),

            {
                'count': 1,
                'limit': 1,
                'time_left': 0,
            })

    @override_settings(LOGIN_LIMIT_RATE='1/s')
    def test_get_usage_count_with_cache_get_error(self) -> None:
        """Testing get_usage_count with cache.get() error"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.spy_on(cache.get, op=kgb.SpyOpRaise(Exception('Oh no')))

        # First call will always have count=1, no matter what.
        with self.assertLogs() as logs:
            self.assertEqual(
                get_usage_count(request=request,
                                increment=True),

                {
                    'count': 0,
                    'limit': 1,
                    'time_left': 0,
                })

        self.assertEqual(len(logs.output), 1)
        self.assertRegex(
            logs.output[0],
            re.compile(
                r'^ERROR:djblets\.auth\.ratelimit:Failed to fetch rate limit '
                r'cache key "example\.com:login-ratelimit:1/\d+"\. Rate limit '
                r'checks are currently unreliable\. Is the cache server '
                r'down\? Error = Oh no\n'
                r'Traceback.*Exception: Oh no',
                re.S))

    @override_settings(LOGIN_LIMIT_RATE='1/s')
    def test_get_usage_count_with_cache_add_error(self) -> None:
        """Testing get_usage_count with cache.add() error"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.spy_on(cache.add, op=kgb.SpyOpRaise(Exception('Oh no')))

        # First call will always have count=1, no matter what.
        with self.assertLogs() as logs:
            self.assertEqual(
                get_usage_count(request=request,
                                increment=True),

                {
                    'count': 0,
                    'limit': 1,
                    'time_left': 0,
                })

        self.assertEqual(len(logs.output), 1)
        self.assertRegex(
            logs.output[0],
            re.compile(
                r'^ERROR:djblets\.auth\.ratelimit:Failed to set rate limit '
                r'cache key "example\.com:login-ratelimit:1/\d+"\. Rate limit '
                r'checks are currently unreliable\. Is the cache server '
                r'down\? Error = Oh no\n'
                r'Traceback.*Exception: Oh no',
                re.S))

    @override_settings(LOGIN_LIMIT_RATE='1/s')
    def test_get_usage_count_with_cache_set_error(self) -> None:
        """Testing get_usage_count with cache.set() error"""
        request = self.request_factory.get('/')
        request.user = User(pk=1)

        self.spy_on(cache.incr, op=kgb.SpyOpRaise(Exception('Oh no')))

        # Put an item in cache.
        self.assertEqual(
            get_usage_count(request=request,
                            increment=True),

            {
                'count': 0,
                'limit': 1,
                'time_left': 0,
            })

        with self.assertLogs() as logs:
            self.assertEqual(
                get_usage_count(request=request,
                                increment=True),

                {
                    'count': 0,
                    'limit': 1,
                    'time_left': 0,
                })

        self.assertEqual(len(logs.output), 1)
        self.assertRegex(
            logs.output[0],
            re.compile(
                r'^ERROR:djblets\.auth\.ratelimit:Failed to set rate limit '
                r'cache key "example\.com:login-ratelimit:1/\d+"\. Rate limit '
                r'checks are currently unreliable\. Is the cache server '
                r'down\? Error = Oh no\n'
                r'Traceback.*Exception: Oh no',
                re.S))



"""Exceptions for integrations."""

from __future__ import annotations

from typing import Optional

from django.utils.translation import gettext as _


class IntegrationError(Exception):
    """Base class for integration-related errors."""


class IntegrationRegistrationError(IntegrationError):
    """Error registering an integration."""


class IntegrationAlreadyRegisteredError(IntegrationRegistrationError):
    """An integration was already registered."""

    def __init__(
        self,
        integration_id: str,
    ) -> None:
        """Initialize the error.

        Args:
            integration_id (str):
                The ID of the integration that was already registered.
        """
        super().__init__(
            _('The integration ID "%s" was already registered.')
            % integration_id)


class IntegrationNotRegisteredError(IntegrationError):
    """Error indicating that the given integration wasn't registered."""

    def __init__(
        self,
        integration_id: Optional[str],
    ) -> None:
        """Initialize the error.

        Args:
            integration_id (str):
                The ID of the integration that has not been registered.
        """
        if integration_id is None:
            message = _('No integration ID has been provided.')
        else:
            message = (
                _('The integration ID "%s" has not been registered.')
                % integration_id
            )

        super().__init__(message)



"""Utilities for capturing and inspecting database queries.

Version Added:
    3.4
"""

from __future__ import annotations

import traceback
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Iterator, List, Sequence, Type, Union

from django.core.exceptions import EmptyResultSet
from django.db.models import Q, QuerySet, Subquery
from django.db.models.expressions import ExpressionWrapper
from django.db.models.signals import pre_delete
from django.db.models.sql.compiler import (SQLCompiler,
                                           SQLDeleteCompiler,
                                           SQLInsertCompiler,
                                           SQLUpdateCompiler)
from django.db.models.sql.query import Query as SQLQuery
from django.db.models.sql.subqueries import AggregateQuery
from django.utils.tree import Node
from typing_extensions import Literal, TypedDict


class ExecutedQueryType(str, Enum):
    """A type of executed query that can be inspected.

    Version Added:
        3.4
    """

    #: A DELETE query.
    DELETE = 'DELETE'

    #: An INSERT query.
    INSERT = 'INSERT'

    #: A SELECT query.
    SELECT = 'SELECT'

    #: An UPDATE query.
    UPDATE = 'UPDATE'


class ExecutedQueryInfo(TypedDict):
    """Information on an executed query.

    This contains information seen at execution time that can be used for
    inspection of the queries.

    Version Added:
        3.4
    """

    #: The query that was executed.
    query: SQLQuery

    #: The type of result information.
    #:
    #: This is used to distinguish between query and subquery information.
    result_type: Literal['query']

    #: The list of lines of SQL that was executed.
    sql: List[str]

    #: Any subqueries within this query, in the orders found.
    subqueries: List[ExecutedSubQueryInfo]

    #: The lines of traceback showing where the query was executed.
    traceback: List[str]

    #: The type of executed query.
    type: ExecutedQueryType


class ExecutedSubQueryInfo(TypedDict):
    """Information on a subquery within an executed query.

    This contains information seen at execution time that can be used for
    inspection of the queries.

    Version Added:
        3.4
    """

    #: The type of class managing the subquery.
    cls: Type[Union[AggregateQuery, QuerySet, Subquery]]

    #: The instance of the subquery class.
    instance: Union[AggregateQuery, QuerySet, Subquery]

    #: The query that was executed.
    query: SQLQuery

    #: The type of result information.
    #:
    #: This is used to distinguish between query and subquery information.
    result_type: Literal['subquery']

    #: Any subqueries within this query, in the orders found.
    subqueries: List[ExecutedSubQueryInfo]

    #: The type of executed query.
    type: Literal[ExecutedQueryType.SELECT]


@dataclass
class CatchQueriesContext:
    """Context for captured query information.

    This is provided and populated when using :py:func:`catch_queries`.

    Version Added:
        3.4
    """

    #: Information on the queries that were executed.
    executed_queries: Sequence[ExecutedQueryInfo]

    #: A mapping of SQL queries to their Q expressions.
    queries_to_qs: Dict[SQLQuery, Q]


@contextmanager
def catch_queries(
    *,
    _check_subqueries: bool = True,
) -> Iterator[CatchQueriesContext]:
    """Catch queries and provide information for further inspection.

    Any database queries executed during this context will be captured and
    provided in the context. For each query, this will capture:

    1. The type of query.
    2. The :py:class:`SQL Query objects <django.db.models.query.sql.Query>`
    3. The generated SQL statements
    4. Tracebacks showing where the SQL was executed.

    It will also provide a mapping of the Query objects to their Q
    expressions.

    To use this, :pypi:`kgb` must be installed.

    Version Added:
        3.4

    Args:
        _check_subqueries (bool, optional):
            Whether to check subqueries.

            This is internal for compatibility with the old behavior for
            :py:meth:`TestCase.assertQueries()
            <djblets.testing.testcases.assertQueries>` and will be removed in a
            future release without a deprecation period.

    Context:
        CatchQueriesContext:
        The context populated with query information.
    """
    try:
        import kgb
    except ImportError as e:
        raise ImportError(
            'kgb could not be imported. It is required in order to use '
            'djblets.db.query_catcher.catch_queries(). Error: %s'
            % e)

    spy_agency = kgb.SpyAgency()

    executed_queries: List[ExecutedQueryInfo] = []
    queries_to_qs: Dict[SQLQuery, Q] = {}

    # Track Query objects any time a compiler is executing SQL.
    @spy_agency.spy_for(SQLCompiler.execute_sql,
                        owner=SQLCompiler)
    def _sql_compiler_execute_sql(
        _self: SQLCompiler,
        *args,
        **kwargs,
    ) -> Any:
        if isinstance(_self, SQLDeleteCompiler):
            query_type = ExecutedQueryType.DELETE
        elif isinstance(_self, SQLUpdateCompiler):
            query_type = ExecutedQueryType.UPDATE
        else:
            query_type = ExecutedQueryType.SELECT

        query = _self.query
        sql = _serialize_caught_sql(query)

        if sql:
            subqueries: List[ExecutedSubQueryInfo] = []

            if _check_subqueries:
                _scan_subqueries(node=query,
                                 result=subqueries,
                                 queries_to_qs=queries_to_qs,
                                 _check_subqueries=_check_subqueries)

            executed_queries.append({
                'query': query,
                'result_type': 'query',
                'sql': sql,
                'subqueries': subqueries,
                'traceback': traceback.format_stack(),
                'type': query_type,
            })

        return SQLCompiler.execute_sql.call_original(_self, *args, **kwargs)

    @spy_agency.spy_for(SQLInsertCompiler.execute_sql,
                        owner=SQLInsertCompiler)
    def _sql_insert_compiler_execute_sql(
        _self: SQLInsertCompiler,
        *args,
        **kwargs,
    ) -> Any:
        query = _self.query
        sql = _serialize_caught_sql(query)

        if sql:
            executed_queries.append({
                'query': query,
                'result_type': 'query',
                'sql': sql,
                'subqueries': [],
                'traceback': traceback.format_stack(),
                'type': ExecutedQueryType.INSERT,
            })

        return SQLInsertCompiler.execute_sql.call_original(_self, *args,
                                                           **kwargs)

    # Build and track Q() objects any time they're added to a Query.
    @spy_agency.spy_for(SQLQuery.add_q, owner=SQLQuery)
    def _query_add_q(
        _self: SQLQuery,
        q_object: Q,
    ) -> Any:
        try:
            queries_to_qs[_self] &= q_object
        except KeyError:
            queries_to_qs[_self] = q_object

        return SQLQuery.add_q.call_original(_self, q_object)

    # Copy Q() objects any time a Query is cloned.
    @spy_agency.spy_for(SQLQuery.clone, owner=SQLQuery)
    def _query_clone(
        _self: SQLQuery,
        *args,
        **kwargs,
    ) -> Any:
        result = SQLQuery.clone.call_original(_self, *args, **kwargs)

        try:
            queries_to_qs[result] = queries_to_qs[_self]
        except KeyError:
            pass

        return result

    # Set up an explicit pre_delete signal. During deletion, Django
    # attempts to determine if it can fast-delete (which can be done if,
    # amongst other things, signals don't need to be emitted for each
    # object).
    #
    # This can lead to inconsistencies in test runs, if registration of
    # pre_delete or post_delete signals is conditional on that run. We
    # want to ensure a stable query count and information for DELETEs,
    # so we deny fast-deletion by setting up a model-global signal handler
    # during capture of queries.
    def _on_pre_delete(**kwargs):
        pass

    pre_delete.connect(_on_pre_delete)

    # Let the caller execute SQL. Results will be stored in the context.
    try:
        yield CatchQueriesContext(executed_queries=executed_queries,
                                  queries_to_qs=queries_to_qs)
    finally:
        # We no longer need to track anything in the compiler of Query.
        spy_agency.unspy_all()
        pre_delete.disconnect(_on_pre_delete)


def _scan_subqueries(
    *,
    node: Union[Node, SQLQuery],
    result: List[ExecutedSubQueryInfo],
    queries_to_qs: Dict[SQLQuery, Q],
    _check_subqueries: bool,
) -> None:
    """Scan for subqueries in a level of a query tree.

    This recursively walks a query tree, looking for any subqueries and
    recording them in the list within the nearest top-level query or subquery.

    Version Added:
        3.4

    Args:
        node (django.utils.tree.Node or django.db.models.sql.Query):
            The level of the tree to process.

        result (list of dict):
            The list of queries to append any new subquery information to.

        queries_to_qs (dict):
            A dictionary containing recording mappings of queries to Q objects.
    """
    child_subqueries: List[ExecutedSubQueryInfo]
    children: List[Any] = []

    if isinstance(node, SQLQuery):
        if isinstance(node, AggregateQuery):
            inner_query = node.inner_query  # type: ignore
            assert isinstance(inner_query, SQLQuery)

            child_subqueries = []
            _scan_subqueries(node=inner_query,
                             result=child_subqueries,
                             queries_to_qs=queries_to_qs,
                             _check_subqueries=_check_subqueries)

            result.append({
                'cls': type(node),
                'instance': node,
                'query': inner_query,
                'result_type': 'subquery',
                'subqueries': child_subqueries,
                'type': ExecutedQueryType.SELECT,
            })

            # We won't have any children to process, so we'll effectively
            # bail at this point.
        else:
            # Process the annotations.
            children = list(node.annotations.values())

            # Continue on by processing the Q filters within it.
            try:
                children += queries_to_qs[node].children
            except KeyError:
                # There are no Q objects for this query. Nothing to scan
                # through.
                pass
    else:
        children = node.children

    for child in children:
        if child:
            if isinstance(child, tuple):
                child = child[1]

            if isinstance(child, Node):
                _scan_subqueries(node=child,
                                 result=result,
                                 queries_to_qs=queries_to_qs,
                                 _check_subqueries=_check_subqueries)
            elif _check_subqueries:
                if isinstance(child, ExpressionWrapper):
                    # NOTE: As of November 28, 2023, django-stubs doesn't have
                    #       type hints for `ExpressionWrapper.expression`.
                    child = child.expression  # type: ignore

                if isinstance(child, (QuerySet, Subquery)):
                    child_subqueries = []
                    _scan_subqueries(node=child.query,
                                     result=child_subqueries,
                                     queries_to_qs=queries_to_qs,
                                     _check_subqueries=_check_subqueries)

                    result.append({
                        'cls': type(child),
                        'instance': child,
                        'query': child.query,
                        'result_type': 'subquery',
                        'subqueries': child_subqueries,
                        'type': ExecutedQueryType.SELECT,
                    })


def _serialize_caught_sql(
    query: SQLQuery,
) -> List[str]:
    """Serialize a caught SQL Query to SQL.

    Version Added:
        3.4

    Args:
        query (django.db.models.query.sql.Query):
            The SQL query to serialize.

    Returns:
        list of str:
        The list of SQL statements executed.
    """
    sql: List[str]

    # Grab the SQL. from the query. This may fail, and if it does, it
    # represents a query that we've caught that isn't actually going to be
    # executed (likely a component of another).
    try:
        try:
            sql = [str(query)]
        except ValueError:
            # When doing an INSERT OR IGNORE (SQLite), the SQL can be a list
            # rather than a string. Query.__str__ doesn't know how to handle
            # this, and will crash. We'll deal with it ourselves.
            sql_statements = query.sql_with_params()
            assert isinstance(sql_statements, list)

            sql = [
                _sql % _sql_params
                for _sql, _sql_params in sql_statements
            ]
    except EmptyResultSet:
        # This will be skipped.
        sql = []

    return sql



"""Import utilities for registries."""

from __future__ import annotations

from importlib import import_module
from threading import Lock

from django.utils.functional import SimpleLazyObject


#: A lock used to control the importing and creation of registries.
#:
#: Version Added:
#:     5.0
_import_lock = Lock()


def lazy_import_registry(module_path, registry_name, **kwargs):
    """Lazily import and construct a registry on access.

    This is useful for providing registry instances in a Django app's
    :file:`__init__.py` file without needing to import other modules that
    might interfere with the Django initialization process. When accessed
    for the first time, the registry will be imported and constructed.

    This can also speed up the startup process, depending on the complexity
    of registries.

    Importing and creation of registries is thread-safe. If two threads try
    to access a registry for the first time at the same time, they'll both
    receive the same instance.

    Version Changed:
        5.0:
        This is now thread-safe.

    Args:
        module_path (str):
            The import path of the module containing the registry.

        registry_name (str):
            The class name of the registry.

        **kwargs (dict):
            Keyword arguments to pass to the registry's constructor.

    Returns:
        django.utils.functional.SimpleLazyObject:
        A wrapper that will dynamically load and forward on to the registry.
    """
    def _create_registry():
        with _import_lock:
            mod = import_module(module_path)

            return getattr(mod, registry_name)(**kwargs)

    return SimpleLazyObject(_create_registry)



"""Filesystem-related utilities."""

from __future__ import annotations

import os
import sys
from typing import Any, Callable, Optional, cast

from django.core.exceptions import SuspiciousFileOperation
from typing_extensions import Protocol


class _PathModule(Protocol):
    """Protocol representing a path module.

    This helps with typing within these utility functions. It should not
    be used outside of this module, and is subject to change.

    Version Added:
        4.0
    """

    abspath: Callable[[str], str]
    dirname: Callable[[str], str]
    join: Callable[..., str]
    normcase: Callable[[str], str]
    relpath: Callable[[str, str], str]
    sep: str


def is_exe_in_path(
    name: str,
) -> bool:
    """Checks whether an executable is in the user's search path.

    This expects a name without any system-specific executable extension.
    It will append the proper extension as necessary. For example,
    use "myapp" and not "myapp.exe".

    This will return True if the app is in the path, or False otherwise.
    """

    if sys.platform == 'win32' and not name.endswith('.exe'):
        name += ".exe"

    for dir in os.environ['PATH'].split(os.pathsep):
        if os.path.exists(os.path.join(dir, name)):
            return True

    return False


def safe_join(
    base: str,
    *paths: str,
    path_mod: Any = os.path,
    rel_to: Optional[str] = None,
) -> str:
    """Safely join filesystem paths, ensuring the result is within a base path.

    This will join paths and and ensure the resulting path doesn't escape the
    base path, making it safer when including non-static path components.

    The result is always an absolute path, unless ``rel_to`` is provided.

    The implementation is based on Django's internal
    :py:func:`django.utils._os.safe_join` function, with additional
    functionality, optimizations, and typing.

    Version Added:
        4.0

    Args:
        base (str):
            The base path that the remaining paths will be joined to.

        paths (tuple of str):
            The resulting paths to join.

        path_mod (module, optional):
            An explicit path module to use.

            This is useful when building non-native filesystem paths.

        rel_to (str, optional):
            A path that the result will be made relative to.

    Returns:
        str:
        The absolute joined path.

    Raises:
        django.core.exceptions.SuspiciousFileOperation:
            The resulting path was outside of the base path.
    """
    _path_mod = cast(_PathModule, path_mod)

    result_path = _path_mod.abspath(_path_mod.join(base, *paths))
    abs_base_path = _path_mod.abspath(base)

    norm_result_path = _path_mod.normcase(result_path)
    norm_base_path = _path_mod.normcase(abs_base_path)

    # We're checking that the resulting path is somewhere within the base
    # path (rather than being the base path or higher up in the tree
    # somewhere).
    #
    # One of three conditions must be satisfied:
    #
    # 1. The resulting path is the same as the base path.
    #
    # 2. The resulting path is within the base path.
    #
    # 3. The resulting path is the root of the filesystem.
    if (norm_result_path != norm_base_path and
        not norm_result_path.startswith(norm_base_path + _path_mod.sep) and
        _path_mod.dirname(norm_base_path) != norm_base_path):
        raise SuspiciousFileOperation(
            'The joined path (%r) is located outside of the base path (%r).'
            % (result_path, abs_base_path))

    if rel_to:
        result_path = _path_mod.relpath(result_path, rel_to)

    return result_path



"""Unit tests for djblets.util.templatetags.djblets_images."""

from __future__ import annotations

import io
import os
import shutil
import tempfile
from collections import OrderedDict

from django.core.exceptions import SuspiciousFileOperation
from django.core.files.base import File
from django.core.files.storage import FileSystemStorage
from django.core.files.uploadedfile import SimpleUploadedFile
from django.template import Context, Template, TemplateSyntaxError

from PIL import Image
from djblets.testing.testcases import TestCase
from djblets.util.templatetags.djblets_images import thumbnail


class SrcSetTagTests(TestCase):
    """Unit tests for the {% srcset %} template tag."""

    def test_with_valid_sources(self):
        """Testing {% srcset %} with valid sources"""
        t = Template('{% load djblets_images %}'
                     '<img srcset="{% srcset sources %}">')

        self.assertEqual(
            t.render(Context({
                'sources': {
                    '1x': '/path/to/image.png',
                    '2x': '/path/to/image@2x.png',
                    '3x': '/path/to/image@3x.png',
                    '10w': '/path/to/image@10w.png',
                },
            })),
            '<img srcset="'
            '/path/to/image.png 1x, '
            '/path/to/image@2x.png 2x, '
            '/path/to/image@3x.png 3x, '
            '/path/to/image@10w.png 10w">')

    def test_with_invalid_sources(self):
        """Testing {% srcset %} with invalid sources"""
        t = Template('{% load djblets_images %}'
                     '<img srcset="{% srcset sources %}">')

        message = '"ax" is not a valid srcset size descriptor.'

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            t.render(Context({
                'sources': {
                    'ax': '/path/to/image.png',
                },
            }))

    def test_sorts(self):
        """Testing {% srcset %} sorts sources"""
        t = Template('{% load djblets_images %}'
                     '<img srcset="{% srcset sources %}">')

        sources = OrderedDict()
        sources['100w'] = '/path/to/image@100w.png'
        sources['10w'] = '/path/to/image@10w.png'
        sources['3x'] = '/path/to/image@3x.png'
        sources['1x'] = '/path/to/image.png'
        sources['2x'] = '/path/to/image@2x.png'
        sources['20w'] = '/path/to/image@20w.png'
        sources['1.5x'] = '/path/to/image@1.5x.png'
        sources['2w'] = '/path/to/image@2w.png'
        sources['10x'] = '/path/to/image@10x.png'

        self.assertEqual(
            t.render(Context({
                'sources': sources,
            })),
            '<img srcset="'
            '/path/to/image.png 1x, '
            '/path/to/image@1.5x.png 1.5x, '
            '/path/to/image@2x.png 2x, '
            '/path/to/image@3x.png 3x, '
            '/path/to/image@10x.png 10x, '
            '/path/to/image@2w.png 2w, '
            '/path/to/image@10w.png 10w, '
            '/path/to/image@20w.png 20w, '
            '/path/to/image@100w.png 100w">')


class ImageSourceAttrsTagTests(TestCase):
    """Unit tests for the {% image_source_attrs %} template tag."""

    def test_with_valid_sources(self):
        """Testing {% image_source_attrs %} with valid sources"""
        t = Template(
            '{% load djblets_images %}'
            '<img {% image_source_attrs %}\n'
            '      1x /path/to/image.png\n'
            '      2x /path/to/image@2x.png\n'
            '      3x /path/to/image@3x.png\n'
            '      10w /path/to/image@10w.png\n'
            '     {% end_image_source_attrs %}>')

        self.assertEqual(
            t.render(Context()),
            '<img src="/path/to/image.png" srcset="'
            '/path/to/image.png 1x, '
            '/path/to/image@2x.png 2x, '
            '/path/to/image@3x.png 3x, '
            '/path/to/image@10w.png 10w">')

    def test_with_invalid_sources(self):
        """Testing {% image_source_attrs %} with invalid sources definition"""
        t = Template(
            '{% load djblets_images %}'
            '<img {% image_source_attrs %}\n'
            '      /path/to/image.png\n'
            '      2x /path/to/image@2x.png\n'
            '     {% end_image_source_attrs %}>')

        message = (
            'The source definition passed to {% image_source_attrs %} is '
            'not structured correctly. Make sure that there is one source '
            'definition per line and that it contains a descriptor and a '
            'URL.'
        )

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            t.render(Context())

    def test_with_no_1x(self):
        """Testing {% image_source_attrs %} with no 1x source"""
        t = Template(
            '{% load djblets_images %}'
            '<img {% image_source_attrs %}\n'
            '      2x /path/to/image@2x.png\n'
            '      3x /path/to/image@3x.png\n'
            '     {% end_image_source_attrs %}>')

        message = (
            'The source definition passed to {% image_source_attr %} must '
            'contain a "1x" descriptor.'
        )

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            t.render(Context())


class ThumbnailTests(TestCase):
    """Unit tests for the {% thumbnail %} template tag.

    Version Added:
        5.0
    """

    #: The root path for the test storage.
    _storage_root: str

    #: The storage instance.
    storage: FileSystemStorage

    def setUp(self) -> None:
        """Set up the test data for a single test run.

        This will create the storage root and instance for the thumbnail
        operations.
        """
        super().setUp()

        self._storage_root = tempfile.mkdtemp()
        self.storage = FileSystemStorage(location=self._storage_root)

    def tearDown(self) -> None:
        """Tear down test data after a single test run.

        This will erase the storage root and clear up the storage instance.
        """
        shutil.rmtree(self._storage_root)
        self.storage = None  # type: ignore

    def test_with_non_image(self) -> None:
        """Testing {% thumbnail %} with non-image file"""
        self.assertEqual(
            thumbnail(SimpleUploadedFile('test.txt', b'test'),
                      storage=self.storage),
            '/media/test.txt')

    def test_with_file_path(self) -> None:
        """Testing {% thumbnail %} with file path"""
        self._build_png_file('sample-wide.png')

        self.assertEqual(
            thumbnail('sample-wide.png',
                      storage=self.storage),
            '/media/sample-wide_400x100.png')

        self._check_image_size(name='sample-wide_400x100.png',
                               expected_width=150,
                               expected_height=100)

    def test_with_file_path_no_storage(self) -> None:
        """Testing {% thumbnail %} with file path and no storage"""
        message = (
            'A file storage backend could not be found for the provided '
            'file.'
        )

        with self.assertRaisesMessage(ValueError, message):
            thumbnail('sample-wide.png')

    def test_with_file_path_outside_storage(self) -> None:
        """Testing {% thumbnail %} with file path outside storage"""
        with self.assertRaises(SuspiciousFileOperation):
            thumbnail('/sample-wide.png',
                      storage=self.storage)

    def test_with_file_from_field(self) -> None:
        """Testing {% thumbnail %} with file from FileField"""
        # This is close enough. We just care about the storage attribute.
        file = self._build_png_file('sample-wide.png')
        setattr(file, 'storage', self.storage)

        self.assertEqual(
            thumbnail(file),
            '/media/sample-wide_400x100.png')

        self._check_image_size(name='sample-wide_400x100.png',
                               expected_width=150,
                               expected_height=100)

    def test_with_file_no_filename(self) -> None:
        """Testing {% thumbnail %} with File without a filename"""
        message = 'The provided file does not have a filename set.'

        with self.assertRaisesMessage(ValueError, message):
            thumbnail(File(name=None,
                           file=None))

    def test_with_image_defaults(self) -> None:
        """Testing {% thumbnail %} with image file and defaults"""
        self.assertEqual(
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage),
            '/media/sample-wide_400x100.png')

        self._check_image_size(name='sample-wide_400x100.png',
                               expected_width=150,
                               expected_height=100)

    def test_with_size_string(self) -> None:
        """Testing {% thumbnail %} with size as string"""
        self.assertEqual(
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size='150x180'),
            '/media/sample-wide_150x180.png')

        self._check_image_size(name='sample-wide_150x180.png',
                               expected_width=150,
                               expected_height=100)

    def test_with_size_string_invalid(self) -> None:
        """Testing {% thumbnail %} with size as invalid string"""
        message = "Thumbnail size '150x180xA' is not valid."

        with self.assertRaisesMessage(ValueError, message):
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size='150x180xA')

    def test_with_size_tuple(self) -> None:
        """Testing {% thumbnail %} with size as tuple"""
        self.assertEqual(
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size=(300, 250)),
            '/media/sample-wide_300x250.png')

        self._check_image_size(name='sample-wide_300x250.png',
                               expected_width=300,
                               expected_height=200)

    def test_with_size_tuple_width_only(self) -> None:
        """Testing {% thumbnail %} with size as tuple and width only"""
        self.assertEqual(
            thumbnail(self._build_png_file('sample-tall.png'),
                      storage=self.storage,
                      size=(300, None)),
            '/media/sample-tall_300.png')

        self._check_image_size(name='sample-tall_300.png',
                               expected_width=300,
                               expected_height=450)

    def test_with_size_tuple_height_only(self) -> None:
        """Testing {% thumbnail %} with size as tuple and height only"""
        self.assertEqual(
            thumbnail(self._build_png_file('sample-tall.png'),
                      storage=self.storage,
                      size=(None, 200)),
            '/media/sample-tall_x200.png')

        self._check_image_size(name='sample-tall_x200.png',
                               expected_width=133,
                               expected_height=200)

    def test_with_size_tuple_invalid(self) -> None:
        """Testing {% thumbnail %} with size as invalid tuple"""
        message = "Thumbnail size (None, None) is not valid."

        with self.assertRaisesMessage(ValueError, message):
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size=(None, None))

    def test_with_size_invalid(self) -> None:
        """Testing {% thumbnail %} with size as invalid type"""
        message = "Thumbnail size 300 is not valid."

        with self.assertRaisesMessage(ValueError, message):
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size=300)  # type: ignore

    def test_with_existing(self) -> None:
        """Testing {% thumbnail %} with existing thumbnail in storage"""
        path = os.path.join(self._storage_root, 'sample-wide_300x250.png')

        with open(path, 'wb') as fp:
            fp.write(b'...\n')

        self.assertEqual(
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size=(300, 250)),
            '/media/sample-wide_300x250.png')

        with open(path, 'rb') as fp:
            self.assertEqual(fp.read(), b'...\n')

    def test_with_existing_and_create_if_missing_false(self) -> None:
        """Testing {% thumbnail %} with existing thumbnail in storage and
        create_if_missing_false"""
        path = os.path.join(self._storage_root, 'sample-wide_300x250.png')

        with open(path, 'wb') as fp:
            fp.write(b'...\n')

        self.assertEqual(
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size=(300, 250),
                      create_if_missing=False),
            '/media/sample-wide_300x250.png')

        with open(path, 'rb') as fp:
            self.assertEqual(fp.read(), b'...\n')

    def test_with_missing_and_create_if_missing_false(self) -> None:
        """Testing {% thumbnail %} with missing thumbnail in storage and
        create_if_missing_false"""
        self.assertIsNone(
            thumbnail(self._build_png_file('sample-wide.png'),
                      storage=self.storage,
                      size=(300, 250),
                      create_if_missing=False))

        self.assertFalse(os.path.exists(
            os.path.join(self._storage_root, 'sample-wide_300x250.png')))

    def _build_png_file(
        self,
        name: str,
    ) -> File:
        """Create a PNG file and return an instance to it.

        Args:
            name (str):
                The name of the testdata image file to copy and return.

        Returns:
            django.core.files.base.File:
            The file instance for that file.
        """
        src_filename = os.path.abspath(os.path.join(
            __file__, '..', 'testdata', name))
        dest_filename = os.path.abspath(os.path.join(
            self._storage_root, name))

        shutil.copy(src_filename, dest_filename)

        return File(name=name,
                    file=None)

    def _check_image_size(
        self,
        *,
        name: str,
        expected_width: int,
        expected_height: int,
    ) -> None:
        """Check the resulting size of a thumbnail image.

        Args:
            name (str):
                The name of the thumbnail file.

            expected_width (int):
                The expected width of the thumbnail.

            expected_height (int):
                The expected height of the thumbnail.

        Raises:
            AssertionError:
                One of the expectations failed.
        """
        path = os.path.join(self._storage_root, name)
        self.assertTrue(os.path.exists(path))

        with open(path, 'rb') as fp:
            image = Image.open(io.BytesIO(fp.read()))

        self.assertEqual(image.width, expected_width)
        self.assertEqual(image.height, expected_height)



"""Utilities for working with functional code.

Version Added:
    4.0
"""

from __future__ import annotations

import itertools
import re
from collections import deque
from typing import Any, AnyStr, Iterable, cast

from django.utils.functional import SimpleLazyObject


def iterable_len(
    iterable: Iterable[Any],
) -> int:
    """Return the length of an iterable.

    This is an efficient way of counting the total number of items in an
    iterable.

    Note that this requires that the iterable is finite in length, and does not
    result in an infinite loop.

    Version Added:
        4.0

    Args:
        iterable (iterable):
            The iterable of which to compute the length.

    Returns:
        int:
        The length of the iterable.
    """
    if hasattr(iterable, '__len__'):
        return len(iterable)  # type: ignore

    # Benchmarks tested across Python 3.7-3.11 show that this is far more
    # efficient than alternative methods (summing and deque + enumerate).
    counter = itertools.count()
    deque(zip(iterable, counter), maxlen=0)

    return next(counter)


def lazy_re_compile(
    regex: AnyStr,
    flags: int = 0,
) -> re.Pattern:
    """Return a lazily-compiled regular expression.

    This will be compiled the first time it's accessed, helping to reduce
    startup time when loading modules or classes that define regular
    expressions.

    Version Added:
        4.0

    Args:
        regex (bytes or str):
            The regular expression to compile.

        flags (int, optional):
            Any flags to pass when compiling the regular expression.

    Returns:
        django.utils.functional.SimpleLazyObject:
        A lazily-resolved regular expression.
    """

    return cast(re.Pattern, SimpleLazyObject(lambda: re.compile(regex, flags)))



"""A builder for static media files.

Version Added:
    5.0
"""

from __future__ import annotations

import inspect
import json
import os
import re
import sys
from pathlib import Path
from typing import (Dict, Optional, Sequence, TYPE_CHECKING, TypeVar,
                    Union, cast)

from django.core.management import call_command
from django.utils.translation import gettext as _
from typing_extensions import Self, TypeAlias

import djblets
from djblets.extensions.errors import ExtensionPackagingError
from djblets.pipeline.settings import build_pipeline_settings
from djblets.util.filesystem import is_exe_in_path

if TYPE_CHECKING:
    from djblets.extensions.extension import (CSSBundleConfigs,
                                              Extension,
                                              JSBundleConfigs)
    from djblets.util.typing import JSONDict

    _BundleConfigsT = TypeVar('_BundleConfigsT',
                              CSSBundleConfigs,
                              JSBundleConfigs)


#: A valid value for a LessCSS variable.
#:
#: Version Added:
#:     5.0
LessCSSValueType: TypeAlias = Union[bool, int, str]


#: A mapping of LessCSS variable names to values.
#:
#: Version Added:
#:     5.0
LessCSSVariables: TypeAlias = Dict[str, LessCSSValueType]


#: A mapping of NPM workspace names to paths.
#:
#: Version Added:
#:     5.0
NPMWorkspaceDirs: TypeAlias = Dict[str, Union[Path, str]]


class StaticMediaBuildContext:
    """Context for performing a static media build for an extension.

    Version Added:
        5.0
    """

    ######################
    # Instance variables #
    ######################

    #: The ID of the package being built.
    #:
    #: Type:
    #:     str
    package_id: str

    #: The absolute path to the root of the source tree.
    #:
    #: Type:
    #:     pathlib.Path
    source_root_dir: Path

    #: The absolute path to the target build directory for the package.
    #:
    #: Type:
    #:     pathlib.Path
    build_dir: Path

    #: The absolute path to the project's NPM workspaces directory.
    #:
    #: Type:
    #:     pathlib.Path
    workspaces_dir: Path

    #: The absolute path to the project's node_modules directory.
    #:
    #: Type:
    #:     pathlib.Path
    node_modules_dir: Path

    #: The absolute path to the source static directory for the extension.
    #:
    #: Type:
    #:     pathlib.Path
    static_dir: Path

    #: The normalized Pipeline CSS bundles to build.
    #:
    #: Type:
    #:     dict
    pipeline_css_bundles: CSSBundleConfigs

    #: The normalized Pipeline JavaScript bundles to build.
    #:
    #: Type:
    #:     dict
    pipeline_js_bundles: JSBundleConfigs

    @classmethod
    def create_for_extension(
        cls,
        *,
        extension_cls: type[Extension],
        source_root_dir: Path,
        build_dir: Optional[Path] = None,
        package_id: Optional[str] = None,
    ) -> Self:
        """Initialize the static media build context for an extension class.

        This will prepare paths, scan the extension, and prepare the static
        media bundles for building with Pipeline.

        Args:
            extension_cls (type):
                The static media extension class.

            source_root_dir (pathlib.Path):
                The root of the source tree.

            build_dir (pathlib.Path, optional):
                The path to the build directory.

                If not provided, this will use a :file:`build` subdirectory off
                of ``source_root_dir``.

            package_id (str, optional):
                An explicit ID for the package.

                If not provided, the extension class name will be used.

        Returns:
            StaticMediaBuildContext:
            The build context.
        """
        # Get the location of the static/ directory within the module in the
        # source tree. We're going to use it to look up static files for
        # input, and as a relative path within the module for the output.
        static_dir = Path(inspect.getfile(extension_cls)).parent / 'static'

        return cls(source_root_dir=source_root_dir,
                   static_dir=static_dir,
                   build_dir=build_dir,
                   package_id=package_id or extension_cls.__name__,
                   css_bundles=extension_cls.css_bundles,
                   js_bundles=extension_cls.js_bundles)

    def __init__(
        self,
        *,
        package_id: str,
        css_bundles: CSSBundleConfigs,
        js_bundles: JSBundleConfigs,
        source_root_dir: Path,
        static_dir: Path,
        build_dir: Optional[Path] = None,
    ) -> None:
        """Initialize the static media build context.

        This will prepare paths, scan the extension, and prepare the static
        media bundles for building with Pipeline.

        Args:
            package_id (str):
                The ID of the package.

            css_bundles (dict):
                The CSS bundles to build.

            js_bundles (dict):
                The JavaScript bundles to build.

            source_root_dir (pathlib.Path):
                The root of the source tree.

            static_dir (pathlib.Path):
                The path to the source directory for static files.

            build_dir (pathlib.Path, optional):
                The path to the build directory.

                If not provided, this will use a :file:`build` subdirectory off
                of ``source_root_dir``.
        """
        source_root_dir = source_root_dir.absolute()

        if build_dir:
            build_dir = build_dir.absolute()

        self.package_id = package_id

        self.build_dir = build_dir or (source_root_dir / 'build')
        self.source_root_dir = source_root_dir
        self.workspaces_dir = source_root_dir / '.npm-workspaces'
        self.node_modules_dir = source_root_dir / 'node_modules'
        self.static_dir = static_dir

        self.pipeline_js_bundles = self._make_pipeline_bundle(
            extension_bundles=js_bundles,
            default_dir=Path('js'),
            ext='.js')
        self.pipeline_css_bundles = self._make_pipeline_bundle(
            extension_bundles=css_bundles,
            default_dir=Path('css'),
            ext='.css')

    def get_npm_workspace_dirs(self) -> NPMWorkspaceDirs:
        """Return NPM workspace directories and symlinks to set up.

        Subclasses can override this to return additional workspaces to
        include. The parent class should always be called and the results
        included.

        Returns:
            dict:
            A mapping of symlink names to target locations.
        """
        return {
            'djblets': Path(djblets.__file__).parent.absolute(),
        }

    def get_lessc_global_vars(self) -> LessCSSVariables:
        """Return a dictionary of LessCSS global variables and their values.

        This can be implemented by subclasses to provide global variables for
        .less files for processing.

        For backwards-compatibility, the following two variables are provided,
        but should not be used:

        ``DEBUG``:
            Set to false, always.

        ``STATIC_ROOT``:
            Set to ``/static/``, always.

        Subclasses can override this to return custom variables.

        Returns:
            dict:
            A dictionary mapping variable names to values.
        """
        return {
            'DEBUG': False,
            'STATIC_ROOT': '/static/',
        }

    def get_lessc_include_path(self) -> set[str]:
        """Return the include path for LessCSS imports.

        By default, this will include the parent directory of every path in
        ``settings.STATICFILES_DIRS``, plus the static directory of the
        extension.

        Returns:
            set of str:
            The de-duplicated set of include paths.
        """
        from django.conf import settings

        less_include: set[str] = set()

        for staticfile_dir in settings.STATICFILES_DIRS:
            if isinstance(staticfile_dir, tuple):
                staticfile_dir = staticfile_dir[1]

            dirname = os.path.dirname(staticfile_dir)
            less_include.add(dirname)

            if os.path.basename(dirname) == 'static':
                # This code is here for compatibility reasons. With a proper
                # Babel setup, .less files should be able to reference others
                # relative to static/ directories. However, some paths may
                # still prefix /static/ to any paths, and we don't want those
                # to break. So, add the parent of this static/ directory
                # in order to allow those lookups to continue working.
                dirname = os.path.dirname(dirname)

            less_include.add(dirname)

        less_include.add(str(self.node_modules_dir))

        return less_include

    def _make_pipeline_bundle(
        self,
        *,
        extension_bundles: _BundleConfigsT,
        default_dir: Path,
        ext: str,
    ) -> _BundleConfigsT:
        """Return a new Pipeline bundle based on an extension's bundle.

        Args:
            extension_bundles (dict):
                The bundles defined on the extension.

            default_dir (pathlib.Path):
                The default directory for the source files, for any bundles
                that don't specify an explicit output filename.

            ext (str):
                The default extension for the source files, for any bundles
                that don't specify an explicit output filename.

        Returns:
            dict:
            The new extension bundle.
        """
        return cast('_BundleConfigsT', {
            name: {
                'source_filenames': list(bundle['source_filenames']),
                'output_filename': (
                    bundle.get('output_filename') or
                    str(default_dir / f'{name}.min{ext}')
                ),
                **{
                    key: bundle.get(key)
                    for key in ('compiler_options',
                                'extra_context',
                                'variant')
                    if key in bundle
                }
            }
            for name, bundle in extension_bundles.items()
        })


class StaticMediaBuilder:
    """A builder for an extension's static media files.

    This is responsible for setting up the build process for CSS and
    JavaScript files for an extension. It will configure Django and Pipeline
    for the build process, collect any information needed to set up the
    build tree (which can be customized via subclasses), and then perform
    the build.

    Version Added:
        5.0
    """

    ######################
    # Instance variables #
    ######################

    #: The build context used for this build.
    #:
    #: Type:
    #:     type
    build_context: StaticMediaBuildContext

    #: Whether NPM support has been checked.
    #:
    #: Type:
    #:     bool
    _checked_npm: bool

    def __init__(
        self,
        *,
        build_context: StaticMediaBuildContext,
    ) -> None:
        """Initialize the builder.

        Args:
            build_context (StaticMediaBuildContext):
                The context used for the build.
        """
        self.build_context = build_context
        self._checked_npm = False

    def prepare_build(self) -> None:
        """Prepare the initial steps for the build.

        This will set up the initial directories and Django environment for
        the build process.
        """
        self.ensure_build_files()
        self.prepare_django_env()

    def ensure_build_files(self) -> None:
        """Set up the build tree and configuration files.

        This will set up the NPM workspaces and a starting
        :file:`package.json` file that can be used to manage the build.

        If there are any changes to the :file:`package.json` in the tree,
        they will be reported to the user.
        """
        build_context = self.build_context
        source_dir = build_context.source_root_dir
        workspaces_dir = build_context.workspaces_dir
        npm_workspace_dirs = build_context.get_npm_workspace_dirs()

        for path in (build_context.build_dir,
                     workspaces_dir):
            if not path.exists():
                path.mkdir(0o755)

        for (npm_workspace_name,
             npm_workspace_dir) in npm_workspace_dirs.items():
            npm_workspace_symlink = workspaces_dir / npm_workspace_name

            if npm_workspace_symlink.exists():
                if npm_workspace_symlink.is_symlink():
                    npm_workspace_symlink.unlink()
                else:
                    sys.stderr.write(
                        _('Cannot symlink %(workspace_name)s to '
                          '%(workspace_target)s. This may cause problems '
                          'building your extension. Check if you can remove '
                          'this path in your tree and try again.')
                        % {
                            'workspace_name': npm_workspace_symlink,
                            'workspace_target': npm_workspace_dir,
                        })
                    continue

            npm_workspace_symlink.symlink_to(npm_workspace_dir)

        # Generate a package.json to install the dependencies for building
        # static media. If there's already one, don't remove it.
        package_json_path = source_dir / 'package.json'
        package_json = self.build_package_json()

        if package_json_path.exists():
            with package_json_path.open('r') as fp:
                existing_package_json = json.load(fp)

            if package_json != existing_package_json:
                # Check if there's anything that may have changed that we
                # would want to set.
                for key, value in package_json.items():
                    if (key not in ('private', 'name') and
                        (key not in existing_package_json or
                         value != existing_package_json[key])):
                        # We found something that may need to be updated.
                        # Complain about it, show the updates, and then
                        # move on.
                        print()
                        print('-' * 75)
                        print()
                        print(f'WARNING: Cowardly refusing to overwrite '
                              f'{package_json_path} with these settings:')
                        print()
                        print(json.dumps(package_json,
                                         indent=2,
                                         sort_keys=True))
                        print()
                        print('You may need to make these changes to your '
                              'own copy manually.')
                        print()
                        print('-' * 75)
                        print()
                        break
        else:
            with package_json_path.open('w') as fp:
                json.dump(package_json,
                          fp,
                          indent=2,
                          sort_keys=True)
                fp.write('\n')

    def prepare_django_env(self) -> None:
        """Set up the Django environment.

        This will set up the settings for both Django and Pipeline to
        control the build process.

        The CSS and JavaScript bundles will be set in Pipeline so that
        Django's build process will run them through Pipeline, outputting to
        the right build directory.
        """
        from django.conf import settings
        from djblets.extensions.staticfiles import PackagingFinder
        from pipeline.conf import settings as pipeline_settings

        build_context = self.build_context
        static_dir = build_context.static_dir

        PackagingFinder.extension_static_dir = str(static_dir)

        settings.STATICFILES_FINDERS = (
            'djblets.extensions.staticfiles.PackagingFinder',
        )
        settings.STORAGES['staticfiles']['BACKEND'] = \
            'djblets.extensions.staticfiles.PackagingCachedFilesStorage'

        settings.STATICFILES_DIRS = [
            *settings.STATICFILES_DIRS,
            str(static_dir),
        ]

        static_root = str(
            build_context.build_dir /
            static_dir.relative_to(Path.cwd())
        )
        settings.STATIC_ROOT = static_root

        settings.INSTALLED_APPS = [
            'django.contrib.staticfiles',
        ]
        settings.CACHES = {
            'default': {
                'BACKEND': 'django.core.cache.backends.locmem.LocMemCache',
            },
        }

        # Update the Pipeline settings.
        pipeline_settings.update(build_pipeline_settings(
            pipeline_enabled=True,
            node_modules_path=str(build_context.node_modules_dir),
            static_root=static_root,
            javascript_bundles=build_context.pipeline_js_bundles,
            stylesheet_bundles=build_context.pipeline_css_bundles,
            use_rollup=True,
            extra_config={
                # Tell djblets.pipeline.compiles.less.LessCompiler not to
                # check for outdated files using its special import script,
                # and to in fact assume all files are outdated so they'll
                # all be rebuilt. This will also prevent access denied
                # issues (due to referenced files not being in the
                # sandboxed tree defined for the static media).
                '_DJBLETS_LESS_ALWAYS_REBUILD': True,

                # Overwrite the arguments from build_pipeline_settings(),
                # and add our own instead.
                'LESS_ARGUMENTS': self._build_lessc_args(),
            }))

    def build_package_json(self) -> JSONDict:
        """Return the contents for a default package.json file.

        Subclasses can override this to return custom content.

        Returns:
            dict:
            The default :file:`package.json` contents.
        """
        build_context = self.build_context
        workspaces_rel_dir = (
            build_context.workspaces_dir
            .relative_to(build_context.source_root_dir)
        )

        return {
            'name': build_context.package_id,
            'private': True,
            'workspaces': [
                f'{workspaces_rel_dir}/*',
            ],
        }

    def install_deps(self) -> None:
        """Install dependencies for building static media.

        This will run :command:`npm install`.
        """
        cwd = os.getcwd()
        os.chdir(self.build_context.source_root_dir)

        try:
            self.npm_install()
        finally:
            os.chdir(cwd)

    def build_static_media(self) -> None:
        """Build static media for the extension."""
        call_command('collectstatic', interactive=False, verbosity=2)

    def prune_source_files(self) -> None:
        """Prune any source files from the package.

        This can be used to produce a Python package without any of the
        source CSS and JavaScript files. Only the final bundles will be
        left behind.
        """
        from django.conf import settings

        build_context = self.build_context
        static_root = Path(settings.STATIC_ROOT)

        # Remove source files from the package, if requested by the
        # packager. This will leave only the CSS and JavaScript bundles.
        self._remove_source_files(build_context.pipeline_css_bundles,
                                  static_root / 'css')
        self._remove_source_files(build_context.pipeline_js_bundles,
                                  static_root / 'js')

    def npm_install(
        self,
        package_spec: Optional[str] = None,
        /,
    ) -> None:
        """Install a package via npm.

        This will first determine if npm is available, and then attempt to
        install the given package.

        Args:
            package_spec (str, optional):
                The package specification (name and optional version range)
                to install.

                If not specified, this will use the default behavior of
                reading :file:`package.json`.

        Raises:
            djblets.extensions.errors.ExtensionPackagingError:
                :command:`npm` could not be found, or there was an error
                installing the package.
        """
        if not self._checked_npm:
            if not is_exe_in_path('npm'):
                raise ExtensionPackagingError(
                    _('Unable to locate npm in the path, which is needed to '
                      'install %s. Static media cannot be built.')
                    % package_spec)

            self._checked_npm = True

        if package_spec:
            # Ensure there's a node_modules directory here (which should be
            # the source directory), so it doesn't install higher up in the
            # directory hierarchy.
            if not os.path.exists('node_modules'):
                os.mkdir('node_modules', 0o755)

            print(_('Installing %(npm_package_spec)s...') % {
                'npm_package_spec': package_spec,
            })
            result = os.system(f'npm install {package_spec}')
        else:
            print(_('Installing npm packages...'))
            result = os.system('npm install')

        if result != 0:
            raise ExtensionPackagingError(_('Installation from npm failed.'))

    def _remove_source_files(
        self,
        pipeline_bundles: _BundleConfigsT,
        media_build_dir: Path,
    ) -> None:
        """Remove all source files, leaving only built bundles.

        Args:
            pipeline_bundles (dict):
                The pipeline bundles to scan for files to remove.

            media_build_dir (str):
                The path to the build directory for the files.
        """
        for root, dirs, files in os.walk(media_build_dir, topdown=False):
            for name in files:
                # A valid file will be represented as one of:
                #
                #     (bundle_name, 'min', stamp, ext)
                #     (bundle_name, 'min', ext)
                #
                # We keep both the pre-stamped and post-stamped versions so
                # that Django's CachedFilesStorage can generate and cache
                # the stamp from the contents of the non-stamped file.
                name_parts = name.split('.')

                if (len(name_parts) < 3 or
                    name_parts[0] not in pipeline_bundles or
                    name_parts[1] != 'min'):
                    # This doesn't appear to be a file representing a bundle,
                    # so we should get rid of it.
                    os.unlink(os.path.join(root, name))

            for name in dirs:
                try:
                    os.rmdir(os.path.join(root, name))
                except Exception:
                    # The directory is probably not empty yet.
                    pass

    def _build_lessc_args(self) -> Sequence[str]:
        """Build the list of arguments for the less compiler.

        This will return a set of arguments that define any global variables
        or include paths for the less compiler. It respects most existing
        arguments already in ``settings.PIPELINE['LESS_ARGUMENTS']``, but
        will filter out any existing include paths and global variables needed
        for packaging.

        Returns:
            list:
            A list of arguments to pass to lessc.
        """
        from django.conf import settings

        build_context = self.build_context

        # Get all the arguments that the application defines, but remove those
        # that we need to override or the ones that aren't compatible with
        # building this media.
        lessc_global_vars = build_context.get_lessc_global_vars()
        exclude_re = re.compile(
            '^--(include-path=|global-var="?(%s)=)'
            % '|'.join(
                re.escape(global_var)
                for global_var in lessc_global_vars.keys()
            ))

        lessc_args = [
            lessc_arg
            for lessc_arg in settings.PIPELINE.get('LESS_ARGUMENTS', [])
            if not exclude_re.match(lessc_arg)
        ]

        lessc_include_paths = sorted(build_context.get_lessc_include_path())

        # Register the include path and any global variables used for
        # building .less files.
        return [
            *lessc_args,
            '--include-path=%s' % os.path.pathsep.join(lessc_include_paths),
            *(
                '--global-var=%s=%s'
                % (key, self._serialize_lessc_value(value))
                for key, value in lessc_global_vars.items()
            )
        ]

    def _serialize_lessc_value(
        self,
        value: LessCSSValueType,
    ) -> str:
        """Serialize a LessCSS-compatible value to a string.

        Args:
            value (LessCSSValueType):
                The value type to serialize.

        Returns:
            str:
            The serialized representation of the value.

        Raises:
            TypeError:
                The provided value passed in is not a valid type.
        """
        if isinstance(value, str):
            return '"%s"' % value.replace('"', r'\"')
        elif isinstance(value, bool):
            if value:
                return 'true'
            else:
                return 'false'
        elif isinstance(value, int):
            return str(value)
        else:
            raise TypeError(
                f'{value!r} is not a valid LessCSS global variable value.'
            )



"""Utilities for working with data structures.

Version Added:
    5.1
"""

from __future__ import annotations

from typing import Any, Sequence


def dict_get_path_value(
    d: dict[str, Any],
    path: Sequence[str],
) -> Any:
    """Return a value from a nested dictionary, given a path of keys.

    This is useful for looking up a value in a nested tree of dictionaries,
    based on a key path.

    Version Added:
        5.1

    Args:
        d (dict):
            The starting dictionary.

        path (list of str):
            The key path.

    Returns:
        object:
        The value at that key.

    Raises:
        KeyError:
            A key in the path could not be found in the dictionary.

        TypeError:
            A value found for part of the key was not a dictionary and could
            not be traversed.
    """
    if not path:
        raise KeyError('The key path cannot be empty.')

    value: Any = d
    i = 0

    try:
        for i, key in enumerate(path):
            # This may raise a KeyError.
            value = value[key]
    except KeyError:
        raise KeyError('Key %r not found.' % path[:i + 1])
    except TypeError:
        # Something in the path was not a dictionary.
        raise TypeError(
            'Key %r is not a dictionary (trying to resolve key %r).'
            % (path[:i], path))

    return value


def dict_set_path_value(
    d: dict[str, Any],
    path: Sequence[str],
    value: Any,
) -> None:
    """Set a value in a nested dictionary, given a path of keys.

    This is useful for setting a value in a nested tree of dictionaries,
    based on a key path.

    If any key in the path points to a non-dictionary value, that value will
    be replaced with a new dictionary.

    Version Added:
        5.1

    Args:
        d (dict):
            The starting dictionary.

        path (list of str):
            The key path.

        value (object):
            The value to set.
    """
    if not path:
        raise KeyError('The key path cannot be empty.')

    for key in path[:-1]:
        child: dict[str, Any]

        try:
            child = d[key]

            if not isinstance(child, dict):
                # Trigger a rebuild of this key.
                raise KeyError
        except KeyError:
            child = {}
            d[key] = child

        d = child

    d[path[-1]] = value



"""Version information for certain Djblets dependencies.

This contains constants that other parts of Djblets and consumers of Djblets
can use to look up information on major dependencies of Djblets.

The contents in this file might change substantially between releases. If
you're going to make use of data from this file, code defensively.
"""

# NOTE: This file may not import other (non-Python) modules! It's used for
#       packaging and may be needed before any dependencies have been
#       installed.

import os
from typing import Dict


###########################################################################
# Python and Django compatibility
###########################################################################

#: The minimum supported version of Python 3.x.
PYTHON_3_MIN_VERSION = (3, 8)

#: A string representation of the minimum supported version of Python 3.x.
PYTHON_3_MIN_VERSION_STR = '%s.%s' % PYTHON_3_MIN_VERSION

#: A dependency version range for Python 3.x.
PYTHON_3_RANGE = ">='%s'" % PYTHON_3_MIN_VERSION_STR


#: The major version of Django we're using for documentation.
django_doc_major_version = '4.2'

#: The major version of Review Board we're using for documentation.
reviewboard_doc_major_version = 'dev'

#: The version range required for Django.
django_version = '~=4.2.15'

###########################################################################
# Python dependencies
###########################################################################

#: All dependencies required to install Djblets.
package_dependencies = {
    'cryptography': '>=41.0.4',
    'Django': django_version,
    'django-pipeline': '~=3.1.0',
    'dnspython': '>=2.3.0',
    'housekeeping': '~=1.1',
    'packaging': '>=23.1',
    'Pillow': '>=6.2',
    'publicsuffixlist': '~=0.10.0',
    'python-dateutil': '>=2.7',
    'pytz': '',
    'typing_extensions': '>=4.4',

    # importlib.metadata compatibility import.
    #
    # 6.6 is equivalent to importlib.metadata in Python 3.12.
    'importlib-metadata': '>=6.6',

    # importlib.resources compatibility import.
    #
    # 5.9 is equivalent to importlib.resources in Python 3.12.
    'importlib-resources': '>=5.9',
}


###########################################################################
# JavaScript dependencies
#
# These are auto-generated when running `npm install --save ...` (if the
# package is not already in node_modules).
#
# To re-generate manually, run: `./contrib/internal/build-npm-deps.py`.
###########################################################################

# Auto-generated Node.js dependencies {


#: Dependencies required for static media building.
frontend_buildkit_npm_dependencies: Dict[str, str] = {
    '@beanbag/frontend-buildkit': '^1.2.0',
    '@beanbag/ink': '^0.5.2',
    '@beanbag/spina': '^3.1.1',
    '@types/jquery': '^3.5.30',
    '@types/underscore': '^1.11.4',
    'backbone': '^1.4.1',
    'jasmine-core': '^5.0.1',
    'jquery': '^3.7.1',
    'jquery-ui': '^1.13.3',
}

#: Dependencies required for static media linting.
lint_npm_dependencies: Dict[str, str] = {
    '@beanbag/eslint-plugin': '^1.0.2',
    'eslint': '^8.29.0',
}


# } Auto-generated Node.js dependencies


#: Node dependencies required to package/develop/test Djblets.
npm_dependencies = {}
npm_dependencies.update(frontend_buildkit_npm_dependencies)
npm_dependencies.update(lint_npm_dependencies)


###########################################################################
# Packaging utilities
###########################################################################

def build_dependency_list(deps, version_prefix=''):
    """Build a list of dependency specifiers from a dependency map.

    This can be used along with :py:data:`package_dependencies`,
    :py:data:`npm_dependencies`, or other dependency dictionaries to build a
    list of dependency specifiers for use on the command line or in
    :file:`setup.py`.

    Args:
        deps (dict):
            A dictionary of dependencies.

    Returns:
        list of unicode:
        A list of dependency specifiers.
    """
    new_deps = []

    for dep_name, dep_details in deps.items():
        if isinstance(dep_details, list):
            new_deps += [
                '%s%s%s; python_version%s'
                % (dep_name, version_prefix, entry['version'], entry['python'])
                for entry in dep_details
            ]
        else:
            new_deps.append('%s%s%s' % (dep_name, version_prefix, dep_details))

    return sorted(new_deps, key=lambda s: s.lower())



"""Unit tests for djblets.util.datastructures.

Version Added:
    3.2
"""

from djblets.testing.testcases import TestCase
from djblets.util.datastructures import (dict_get_path_value,
                                         dict_set_path_value)


class DictGetPathValueTests(TestCase):
    """Unit tests for dict_get_path_value.

    Version Added:
        3.2
    """

    def test_with_single_dict(self):
        """Testing dict_get_path_value with single-level dictionary"""
        d = {
            'a': 1,
        }

        self.assertEqual(dict_get_path_value(d, ['a']), 1)

    def test_with_single_dict_key_not_found(self):
        """Testing dict_get_path_value with single-level dictionary and
        key not found
        """
        d = {
            'a': 1,
        }

        with self.assertRaisesMessage(KeyError, "Key ['b'] not found."):
            dict_get_path_value(d, ['b'])

    def test_with_nested_dict(self):
        """Testing dict_get_path_value with nested dictionaries"""
        d = {
            'a': {
                'b': {
                    'c': 123,
                },
            },
        }

        self.assertEqual(
            dict_get_path_value(d, ['a']),
            {
                'b': {
                    'c': 123,
                },
            })
        self.assertEqual(
            dict_get_path_value(d, ['a', 'b']),
            {
                'c': 123,
            })
        self.assertEqual(
            dict_get_path_value(d, ['a', 'b', 'c']),
            123)

    def test_with_nested_dict_not_found(self):
        """Testing dict_get_path_value with nested dictionaries and
        key not found
        """
        d = {
            'a': {
                'b': {
                    'c': 123,
                },
            },
        }

        with self.assertRaisesMessage(KeyError, "Key ['a', 'd'] not found."):
            dict_get_path_value(d, ['a', 'd'])

        with self.assertRaisesMessage(KeyError,
                                      "Key ['a', 'b', 'd'] not found."):
            dict_get_path_value(d, ['a', 'b', 'd'])

    def test_with_nested_dict_bad_type(self):
        """Testing dict_get_path_value with nested dictionaries and
        non-dict key in path
        """
        d = {
            'a': {
                'b': {
                    'c': 123,
                },
            },
        }

        message = (
            "Key ['a', 'b', 'c'] is not a dictionary (trying to resolve key "
            "['a', 'b', 'c', 'd'])."
        )

        with self.assertRaisesMessage(TypeError, message):
            dict_get_path_value(d, ['a', 'b', 'c', 'd'])

    def test_with_empty_path(self):
        """Testing dict_get_path_value with empty path"""
        with self.assertRaisesMessage(KeyError,
                                      'The key path cannot be empty.'):
            dict_get_path_value({}, [])


class DictSetPathValueTests(TestCase):
    """Unit tests for dict_set_path_value.

    Version Added:
        3.2
    """

    def test_with_single_dict(self):
        """Testing dict_set_path_value with single-level dictionary"""
        d = {}

        dict_set_path_value(d, ['a'], 1)

        self.assertEqual(d, {
            'a': 1,
        })

    def test_with_nested_dict(self):
        """Testing dict_set_path_value with nested dictionaries"""
        d = {}

        dict_set_path_value(d, ['a', 'b', 'c'], 1)

        self.assertEqual(d, {
            'a': {
                'b': {
                    'c': 1,
                },
            },
        })

        # Make sure merging doesn't break anything.
        dict_set_path_value(d, ['a', 'b', 'd'], 'test')
        dict_set_path_value(d, ['a', 'e'], True)

        self.assertEqual(d, {
            'a': {
                'b': {
                    'c': 1,
                    'd': 'test',
                },
                'e': True,
            },
        })

    def test_overrides_non_dicts(self):
        """Testing dict_set_path_value overrides non-dictionary keys"""
        d = {
            'a': 1,
        }

        dict_set_path_value(d, ['a', 'b'], 'test')

        self.assertEqual(d, {
            'a': {
                'b': 'test',
            },
        })

    def test_with_empty_path(self):
        """Testing dict_set_path_value with empty path"""
        with self.assertRaisesMessage(KeyError,
                                      'The key path cannot be empty.'):
            dict_set_path_value({}, [], 123)



"""Unit tests for djblets.http.responses.EventStreamHttpResponse.

Version Added:
    4.0
"""

from __future__ import annotations

from typing import Optional

from django.middleware.gzip import GZipMiddleware
from django.test.client import RequestFactory

from djblets.http.responses import (EventStreamHttpResponse,
                                    EventStreamMessages)
from djblets.testing.testcases import TestCase


class EventStreamHttpResponseTests(TestCase):
    """Unit tests for EventStreamHttpResponse.

    Version Added:
        4.0
    """

    def test_with_generator(self) -> None:
        """Testing EventStreamHttpResponse with generator"""
        response = EventStreamHttpResponse(
            (
                {
                    'data': b'This is message %d.' % i,
                }
                for i in range(5)
            )
        )

        self.assertEqual(
            list(response),
            [
                b'data: This is message 0.\n'
                b'\n',

                b'data: This is message 1.\n'
                b'\n',

                b'data: This is message 2.\n'
                b'\n',

                b'data: This is message 3.\n'
                b'\n',

                b'data: This is message 4.\n'
                b'\n',
            ])

    def test_with_callable(self) -> None:
        """Testing EventStreamHttpResponse with callable"""
        def _gen_messages(
            last_id: Optional[str],
        ) -> EventStreamMessages:
            self.assertIsNone(last_id)

            for i in range(5):
                yield {
                    'data': f'This is message {i}.',
                }

        response = EventStreamHttpResponse(_gen_messages)

        self.assertEqual(
            list(response),
            [
                b'data: This is message 0.\n'
                b'\n',

                b'data: This is message 1.\n'
                b'\n',

                b'data: This is message 2.\n'
                b'\n',

                b'data: This is message 3.\n'
                b'\n',

                b'data: This is message 4.\n'
                b'\n',
            ])

    def test_with_all_fields(self) -> None:
        """Testing EventStreamHttpResponse with all field types"""
        response = EventStreamHttpResponse(
            (
                {
                    'data': f'This is message {i}.',
                    'event': 'update',
                    'id': f'id-{i}',
                    'retry_ms': 123,
                }
                for i in range(5)
            )
        )

        self.assertEqual(
            list(response),
            [
                b'id: id-0\n'
                b'event: update\n'
                b'data: This is message 0.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-1\n'
                b'event: update\n'
                b'data: This is message 1.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-2\n'
                b'event: update\n'
                b'data: This is message 2.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-3\n'
                b'event: update\n'
                b'data: This is message 3.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-4\n'
                b'event: update\n'
                b'data: This is message 4.\n'
                b'retry: 123\n'
                b'\n',
            ])

    def test_with_all_blank_id(self) -> None:
        """Testing EventStreamHttpResponse with blank ID"""
        response = EventStreamHttpResponse(
            (
                {
                    'id': '',
                    'data': f'test {i}',
                }
                for i in range(2)
            )
        )

        self.assertEqual(
            list(response),
            [
                b'id\n'
                b'data: test 0\n'
                b'\n',

                b'id\n'
                b'data: test 1\n'
                b'\n',
            ])

    def test_with_all_multiline_data(self) -> None:
        """Testing EventStreamHttpResponse with multi-line data"""
        response = EventStreamHttpResponse(
            (
                {
                    'data': ''.join(
                        f'this is line {j + 1}\n'
                        for j in range(i + 1)
                    )
                }
                for i in range(3)
            )
        )

        self.assertEqual(
            list(response),
            [
                b'data: this is line 1\n'
                b'\n',

                b'data: this is line 1\n'
                b'data: this is line 2\n'
                b'\n',

                b'data: this is line 1\n'
                b'data: this is line 2\n'
                b'data: this is line 3\n'
                b'\n',
            ])

    def test_with_last_event_id(self) -> None:
        """Testing EventStreamHttpResponse with Last-Event-ID"""
        def _gen_messages(
            last_id: Optional[str],
        ) -> EventStreamMessages:
            assert last_id == '3'

            for i in range(int(last_id), 5):
                yield {
                    'data': f'This is message {i}.',
                }

        request = RequestFactory().get('/', HTTP_LAST_EVENT_ID='3')

        response = EventStreamHttpResponse(_gen_messages,
                                           request=request)

        self.assertEqual(
            list(response),
            [
                b'data: This is message 3.\n'
                b'\n',

                b'data: This is message 4.\n'
                b'\n',
            ])

    def test_with_gzip_middleware(self) -> None:
        """Testing EventStreamHttpResponse with GZipMiddleware"""

        # NOTE: django-stubs hard-codes HttpResponse as the result of the
        #       lambda, so we have to ignore the typing.
        middleware = GZipMiddleware(
            lambda request: EventStreamHttpResponse(  # type: ignore
                (
                    {
                        'data': f'This is message {i}.',
                        'event': 'update',
                        'id': f'id-{i}',
                        'retry_ms': 123,
                    }
                    for i in range(5)
                ),
                request=request))

        response = middleware(
            RequestFactory().get('/', HTTP_LAST_EVENT_ID='3'))

        self.assertEqual(
            list(response),
            [
                b'id: id-0\n'
                b'event: update\n'
                b'data: This is message 0.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-1\n'
                b'event: update\n'
                b'data: This is message 1.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-2\n'
                b'event: update\n'
                b'data: This is message 2.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-3\n'
                b'event: update\n'
                b'data: This is message 3.\n'
                b'retry: 123\n'
                b'\n',

                b'id: id-4\n'
                b'event: update\n'
                b'data: This is message 4.\n'
                b'retry: 123\n'
                b'\n',
            ])



from __future__ import annotations

from typing import Set

from djblets.extensions.hooks import ExtensionHookPoint
from djblets.integrations.hooks import BaseIntegrationHook
from djblets.integrations.integration import Integration
from djblets.integrations.manager import IntegrationManager
from djblets.integrations.tests.models import IntegrationConfig
from djblets.integrations.tests.testcases import IntegrationsTestCase


class DummyIntegrationHook(BaseIntegrationHook, metaclass=ExtensionHookPoint):
    def __init__(self, manager, *args, **kwargs):
        self.manager = manager

        super().__init__(*args, **kwargs)

    def get_integration_manager(self):
        return self.manager


class DummyIntegration(Integration):
    def initialize(self):
        pass


class IntegrationHookTests(IntegrationsTestCase):
    """Unit tests for djblets.integrations.hooks.BaseIntegrationHook."""

    def setUp(self) -> None:
        super().setUp()

        self.manager = IntegrationManager(IntegrationConfig)

        # This will make this class a hook owner.
        self.hooks: Set[DummyIntegrationHook] = set()

        self.hook = DummyIntegrationHook(self.manager, self, DummyIntegration)

    def test_initialize(self) -> None:
        """Testing BaseIntegrationHook.initialize"""
        self.assertIn(DummyIntegration.integration_id,
                      self.manager._integration_classes)

    def test_shutdown(self) -> None:
        """Testing BaseIntegrationHook.initialize"""
        self.hook.disable_hook()

        self.assertNotIn(DummyIntegration.integration_id,
                         self.manager._integration_classes)
